from typing import Any, Iterator, List, Optional, Set, Tuple, Union

import dask.base
import dask.dataframe
import pandas

from ..storage import Storage
from ..structures.core import Spec, StructureFamily
from ..structures.table import TableStructure
from ..type_aliases import JSON
from .array import ArrayAdapter


class TableAdapter:
    """
    Wrap a dataframe-like object in an interface that Tiled can serve.

    Examples
    --------

    >>> df = pandas.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    >>> DataFrameAdapter.from_pandas(df, npartitions=1)

    """

    structure_family = StructureFamily.table
    supported_storage: Set[type[Storage]] = set()

    def __init__(
        self,
        partitions: Union[dask.dataframe.DataFrame, pandas.DataFrame],
        structure: TableStructure,
        *,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> None:
        """

        Parameters
        ----------
        partitions :
        structure :
        metadata :
        specs :
        """
        self._metadata = metadata or {}
        self._partitions = list(partitions)
        self._structure = structure
        self.specs = specs or []

    @classmethod
    def from_pandas(
        cls,
        *args: Any,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
        npartitions: int = 1,
        **kwargs: Any,
    ) -> "TableAdapter":
        """

        Parameters
        ----------
        args :
        metadata :
        specs :
        npartitions :
        kwargs :

        Returns
        -------

        """
        ddf = dask.dataframe.from_pandas(*args, npartitions=npartitions, **kwargs)
        if specs is None:
            specs = [Spec("dataframe")]
        return cls.from_dask_dataframe(ddf, metadata=metadata, specs=specs)

    @classmethod
    def from_dict(
        cls,
        *args: Any,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
        npartitions: int = 1,
        **kwargs: Any,
    ) -> "TableAdapter":
        """

        Parameters
        ----------
        args :
        metadata :
        specs :
        npartitions :
        kwargs :

        Returns
        -------

        """
        ddf = dask.dataframe.from_dict(*args, npartitions=npartitions, **kwargs)
        if specs is None:
            specs = [Spec("dataframe")]
        return cls.from_dask_dataframe(ddf, metadata=metadata, specs=specs)

    @classmethod
    def from_dask_dataframe(
        cls,
        ddf: dask.dataframe.DataFrame,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> "TableAdapter":
        """

        Parameters
        ----------
        ddf :
        metadata :
        specs :

        Returns
        -------

        """
        structure = TableStructure.from_dask_dataframe(ddf)
        if specs is None:
            specs = [Spec("dataframe")]
        return cls(
            ddf.partitions,
            structure,
            metadata=metadata,
            specs=specs,
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._structure.columns!r})"

    def __getitem__(self, key: str) -> ArrayAdapter:
        # Must compute to determine shape
        array = self.read([key])[key].values

        return ArrayAdapter.from_array(array)

    def get(self, key: str) -> Union[ArrayAdapter, None]:
        if key not in self.structure().columns:
            return None
        return self[key]

    def items(self) -> Iterator[Tuple[str, ArrayAdapter]]:
        yield from ((key, self[key]) for key in self._structure.columns)

    def metadata(self) -> JSON:
        return self._metadata

    def structure(self) -> TableStructure:
        return self._structure

    def read(
        self, fields: Optional[List[str]] = None
    ) -> Union[dask.dataframe.DataFrame, pandas.DataFrame]:
        """

        Parameters
        ----------
        fields :

        Returns
        -------

        """
        if any(p is None for p in self._partitions):
            raise ValueError("Not all partitions have been stored.")
        if isinstance(self._partitions[0], dask.dataframe.DataFrame):
            if fields is not None:
                ddf = dask.dataframe.concat(
                    [p[fields] for p in self._partitions], axis=0
                )
            else:
                ddf = dask.dataframe.concat(self._partitions, axis=0)
            return ddf.compute()
        df = pandas.concat(self._partitions, axis=0)
        if fields is not None:
            df = df[fields]
        return df

    def read_partition(
        self,
        partition: int,
        fields: Optional[List[str]] = None,
    ) -> Union[pandas.DataFrame, dask.dataframe.DataFrame]:
        """

        Parameters
        ----------
        partition :
        fields :

        Returns
        -------

        """
        df = self._partitions[partition]
        if df is None:
            raise RuntimeError(f"Partition {partition} has not been stored yet.")
        if fields is not None:
            df = df[fields]
        if isinstance(df, dask.dataframe.DataFrame):
            return df.compute()
        return partition


DataFrameAdapter = TableAdapter



from abc import ABC
from dataclasses import dataclass
from typing import Optional

from fastapi import Request


@dataclass
class UserSessionState:
    """Data transfer class to communicate custom session state information."""

    user_name: str
    state: dict = None


class InternalAuthenticator(ABC):
    def authenticate(self, username: str, password: str) -> Optional[UserSessionState]:
        raise NotImplementedError


class ExternalAuthenticator(ABC):
    def authenticate(self, request: Request) -> Optional[UserSessionState]:
        raise NotImplementedError



from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel


class AboutAuthenticationProvider(BaseModel):
    provider: str
    # "password" is a back-compat alias for "internal"; "internal should be preferred.
    mode: Literal["internal", "external", "password"]
    links: Dict[str, str]
    confirmation_message: Optional[str] = None


class AboutAuthenticationLinks(BaseModel):
    whoami: str
    apikey: str
    refresh_session: str
    revoke_session: str
    logout: str


class AboutAuthentication(BaseModel):
    required: bool
    providers: List[AboutAuthenticationProvider]
    links: Optional[AboutAuthenticationLinks] = None


class About(BaseModel):
    api_version: int
    library_version: str
    formats: Dict[str, List[str]]
    aliases: Dict[str, Dict[str, List[str]]]
    queries: List[str]
    authentication: AboutAuthentication
    links: Dict[str, str]
    meta: Dict[str, Any]



import dataclasses
import enum
from typing import Generic, List, Optional, TypeVar

from .core import StructureFamily


class Management(str, enum.Enum):
    external = "external"
    immutable = "immutable"
    locked = "locked"
    writable = "writable"


@dataclasses.dataclass
class Asset:
    data_uri: str
    is_directory: bool
    parameter: Optional[str]
    num: Optional[int] = None
    id: Optional[int] = None


StructureT = TypeVar("StructureT")


@dataclasses.dataclass
class DataSource(Generic[StructureT]):
    structure_family: StructureFamily
    structure: StructureT
    id: Optional[int] = None
    mimetype: Optional[str] = None
    parameters: dict = dataclasses.field(default_factory=dict)
    assets: List[Asset] = dataclasses.field(default_factory=list)
    management: Management = Management.writable

    @classmethod
    def from_json(cls, d):
        d = d.copy()
        assets = [Asset(**a) for a in d.pop("assets")]
        return cls(assets=assets, **d)



import itertools
import warnings

import httpx

from ..structures.core import StructureFamily
from ..structures.data_source import DataSource, Management
from .base import BaseClient
from .utils import ClientError


def copy(
    source: BaseClient,
    dest: BaseClient,
    on_conflict: str = "error",
):
    """
    Copy data from one Tiled instance to another.

    Parameters
    ----------
    source : tiled node
    dest : tiled node
    on_conflict : str, default 'error', other options 'warn', 'skip'

    Examples
    --------

    Connect to two instances and copy data.

    >>> from tiled.client import from_uri
    >>> from tiled.client.sync import copy
    >>> a = from_uri("http://localhost:8000", api_key="secret")
    >>> b = from_uri("http://localhost:9000", api_key="secret")
    >>> copy(a, b)


    Copy select data.

    >>> copy(a.items().head(), b)
    >>> copy(a.search(...), b)

    Copy and ignore duplicates.

    >>> copy(a, b, on_conflict = 'skip')
    """
    if hasattr(source, "structure_family"):
        # looks like a client object
        _DISPATCH[source.structure_family](
            source.include_data_sources(), dest, on_conflict
        )
    else:
        _DISPATCH[StructureFamily.container](dict(source), dest, on_conflict)


def _copy_array(source, dest, on_conflict):
    num_blocks = (range(len(n)) for n in source.chunks)
    # Loop over each block index --- e.g. (0, 0), (0, 1), (0, 2) ....
    for block in itertools.product(*num_blocks):
        array = source.read_block(block)
        dest.write_block(array, block)


def _copy_awkward(source, dest, on_conflict):
    import awkward

    array = source.read()
    _form, _length, container = awkward.to_buffers(array)
    dest.write(container)


def _copy_sparse(source, dest, on_conflict):
    num_blocks = (range(len(n)) for n in source.chunks)
    # Loop over each block index --- e.g. (0, 0), (0, 1), (0, 2) ....
    for block in itertools.product(*num_blocks):
        array = source.read_block(block)
        dest.write_block(array.coords, array.data, block)


def _copy_table(source, dest, on_conflict):
    for partition in range(source.structure().npartitions):
        df = source.read_partition(partition)
        dest.write_partition(df, partition)


def _copy_container(source, dest, on_conflict):
    for key, child_node in source.items():
        original_data_sources = child_node.include_data_sources().data_sources()
        num_data_sources = len(original_data_sources)
        if num_data_sources == 0:
            # A container with no data sources is just an organizational
            # entity in the database.
            if child_node.structure_family == StructureFamily.container:
                data_sources = []
            else:
                raise ValueError(
                    f"Unable to copy {child_node} which is a "
                    f"{child_node.structure_family} but has no data sources."
                )
        elif num_data_sources == 1:
            (original_data_source,) = original_data_sources
            if original_data_source.management == Management.external:
                data_sources = [original_data_source]
            else:
                if child_node.structure_family == StructureFamily.container:
                    data_sources = []
                else:
                    data_sources = [
                        DataSource(
                            management=original_data_source.management,
                            mimetype=original_data_source.mimetype,
                            structure_family=original_data_source.structure_family,
                            structure=original_data_source.structure,
                        )
                    ]
        else:
            # As of this writing this is impossible, but we anticipate that
            # it may be added someday.
            raise NotImplementedError(
                "Multiple Data Sources in one Node is not supported."
            )
        try:
            node = dest.new(
                key=key,
                structure_family=child_node.structure_family,
                data_sources=data_sources,
                metadata=dict(child_node.metadata),
                specs=child_node.specs,
            )
        except ClientError as err:
            if (
                on_conflict == "skip" or on_conflict == "warn"
            ) and err.response.status_code == httpx.codes.CONFLICT:
                if on_conflict == "warn":
                    warnings.warn("Skipped existing entry")
                continue
            else:
                raise err
        if (
            original_data_sources
            and (original_data_sources[0].management != Management.external)
        ) or (
            (child_node.structure_family == StructureFamily.container)
            and (not original_data_sources)
        ):
            _DISPATCH[child_node.structure_family](
                child_node, node, on_conflict=on_conflict
            )


_DISPATCH = {
    StructureFamily.array: _copy_array,
    StructureFamily.awkward: _copy_awkward,
    StructureFamily.container: _copy_container,
    StructureFamily.sparse: _copy_sparse,
    StructureFamily.table: _copy_table,
}



"""Localize all timestamps

Revision ID: d88e91ea03f9
Revises: 13024b8a6b74
Create Date: 2025-02-18 14:15:48.967976

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d88e91ea03f9"
down_revision = "13024b8a6b74"
branch_labels = None
depends_on = None


# mapping table names to list of column given as (name, nullable)
datetime_columns = {
    "principals": [("time_created", True), ("time_updated", True)],
    "identities": [
        ("latest_login", True),
        ("time_created", False),
        ("time_updated", True),
    ],
    "roles": [("time_created", True), ("time_updated", True)],
    "api_keys": [
        ("expiration_time", True),
        ("latest_activity", True),
        ("time_created", False),
        ("time_updated", True),
    ],
    "sessions": [
        ("expiration_time", False),
        ("time_created", False),
        ("time_last_refreshed", True),
        ("time_updated", True),
    ],
    "pending_sessions": [("expiration_time", False)],
}


def upgrade():
    connection = op.get_bind()
    if connection.engine.dialect.name == "sqlite":
        # No action required. SQLAlchemy handles timezones at the application
        # level and does not store anything like +00 or +Z in the database.
        return

    for table, columns in datetime_columns.items():
        for column, nullable in columns:
            # Create a temporary column with localized datetimes.
            # Hardcode nullable=True because the values are not initialized.
            # We will set nullable properly at the end.
            op.add_column(
                table,
                # NOTE: Later it was noticed that server_default was missed here.
                # The following migration (0c705a02954c) fixed this.
                sa.Column(
                    f"{column}_localized", sa.DateTime(timezone=True), nullable=True
                ),
            )

            # Copy date from naive to localized column.
            connection.execute(
                sa.text(
                    f"""
                    UPDATE {table}
                    SET {column}_localized = {column} AT TIME ZONE 'UTC'
                    WHERE {column} IS NOT NULL
                """
                )
            )

            # Drop the original (naive) column.
            op.drop_column(table, column)

            # Rename the new column to the original name, and set nullable to
            # the correct value.
            op.alter_column(
                table, f"{column}_localized", new_column_name=column, nullable=nullable
            )


def downgrade():
    connection = op.get_bind()
    if connection.engine.dialect.name == "sqlite":
        # No action required. SQLAlchemy handles timezones at the application
        # level and does not store anything like +00 or +Z in the database.
        return

    for table, columns in datetime_columns.items():
        for column, nullable in columns:
            # Create a temporary column with naive datetimes.
            # Hardcode nullable=True because the values are not initialized.
            # We will set nullable properly at the end.
            op.add_column(
                table,
                sa.Column(
                    f"{column}_naive", sa.DateTime(timezone=False), nullable=False
                ),
            )

            # Copy date from localized to naive column.
            connection.execute(
                sa.text(
                    f"""
                    UPDATE {table}
                    SET {column}_naive = {column} AT TIME ZONE 'UTC'
                    WHERE {column} IS NOT NULL
                """
                )
            )

            # Drop the original (localized) column.
            op.drop_column(table, column)

            # Rename the new column to the original name, and set nullable to
            # the correct value.
            op.alter_column(
                table, f"{column}_naive", new_column_name=column, nullable=nullable
            )



import logging
import os

from ..queries import AccessBlobFilter
from ..utils import Sentinel, import_object
from .scopes import ALL_SCOPES, PUBLIC_SCOPES

ALL_ACCESS = Sentinel("ALL_ACCESS")
NO_ACCESS = Sentinel("NO_ACCESS")


logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
handler.setLevel("DEBUG")
handler.setFormatter(logging.Formatter("TILED ACCESS POLICY: %(message)s"))
logger.addHandler(handler)

log_level = os.getenv("TILED_ACCESS_POLICY_LOG_LEVEL")
if log_level:
    logger.setLevel(log_level.upper())


class DummyAccessPolicy:
    "Impose no access restrictions."

    async def allowed_scopes(self, node, principal, authn_access_tags, authn_scopes):
        return ALL_SCOPES

    async def filters(self, node, principal, authn_access_tags, authn_scopes, scopes):
        return []


class TagBasedAccessPolicy:
    def __init__(
        self,
        *,
        provider,
        tags_db,
        access_tags_parser,
        scopes=None,
    ):
        self.provider = provider
        self.scopes = scopes if (scopes is not None) else ALL_SCOPES

        access_tags_parser = import_object(access_tags_parser)
        self.access_tags_parser = access_tags_parser.from_uri(tags_db["uri"])
        self.is_tag_defined = self.access_tags_parser.is_tag_defined
        self.get_public_tags = self.access_tags_parser.get_public_tags
        self.get_scopes_from_tag = self.access_tags_parser.get_scopes_from_tag
        self.is_tag_owner = self.access_tags_parser.is_tag_owner
        self.is_tag_public = self.access_tags_parser.is_tag_public
        self.get_tags_from_scope = self.access_tags_parser.get_tags_from_scope

        self.read_scopes = PUBLIC_SCOPES
        self.unremovable_scopes = ["read:metadata", "write:metadata"]
        self.admin_scopes = ["admin:apikeys"]
        self.public_tag = "public".casefold()
        self.invalid_tag_names = [name.casefold() for name in []]

    def _get_id(self, principal):
        for identity in principal.identities:
            if identity.provider == self.provider:
                return identity.id
        else:
            raise ValueError(
                f"Principal {principal} has no identity from provider {self.provider}."
                f"The Principal's identities are: {principal.identities}"
            )

    def _is_admin(self, authn_scopes):
        if all(s in authn_scopes for s in self.admin_scopes):
            return True
        return False

    async def init_node(
        self, principal, authn_access_tags, authn_scopes, access_blob=None
    ):
        if principal.type == "service":
            identifier = str(principal.uuid)
        else:
            identifier = self._get_id(principal)

        if access_blob:
            if len(access_blob) != 1 or "tags" not in access_blob:
                raise ValueError(
                    f"""access_blob must be in the form '{{"tags": ["tag1", "tag2", ...]}}'\n"""
                    f"""Received {access_blob=}"""
                )
            if not access_blob["tags"]:
                if not self._is_admin(authn_scopes):
                    raise ValueError(
                        "Cannot apply empty tag list to node: only Tiled admins can apply an empty tag list."
                    )
            access_tags = set(access_blob["tags"])
            include_public_tag = False
            for tag in access_tags:
                if authn_access_tags is not None:
                    if tag not in authn_access_tags:
                        raise ValueError(
                            f"Cannot apply tag to node: API key is restricted to access tags: {authn_access_tags}."
                        )
                if tag.casefold() == self.public_tag:
                    include_public_tag = True
                    if not self._is_admin(authn_scopes):
                        raise ValueError(
                            "Cannot apply 'public' tag to node: only Tiled admins can apply the 'public' tag."
                        )
                elif not await self.is_tag_defined(tag):
                    raise ValueError(f"Cannot apply tag to node: {tag=} is not defined")
                elif not await self.is_tag_owner(tag, identifier):
                    # admins can ignore the tag ownership check
                    if not self._is_admin(authn_scopes):
                        raise ValueError(
                            f"Cannot apply tag to node: user='{identifier}' is not an owner of {tag=}"
                        )
                elif tag.casefold() in self.invalid_tag_names:
                    raise ValueError(
                        f"Cannot apply tag to node: '{tag}' is not a valid tag name."
                    )

            access_tags_from_policy = {
                tag for tag in access_tags if tag.casefold() != self.public_tag
            }
            if include_public_tag:
                access_tags_from_policy.add(self.public_tag)

            access_blob_from_policy = {"tags": list(access_tags_from_policy)}
            access_blob_modified = access_tags != access_tags_from_policy

            # admin principals are not subject to scope reduction restriction
            if not self._is_admin(authn_scopes):
                # check that the access_blob would not result in invalid scopes for user.
                new_scopes = set()
                for tag in access_tags_from_policy:
                    new_scopes.update(await self.get_scopes_from_tag(tag, identifier))
                if not all(scope in new_scopes for scope in self.unremovable_scopes):
                    raise ValueError(
                        f"Cannot init node with tags: operation does not grant necessary scopes.\n"
                        f"The resulting access_blob would be: {access_blob_from_policy}\n"
                        f"This access_blob does not confer the minimum scopes: {self.unremovable_scopes}"
                    )
        else:
            if authn_access_tags is not None:
                raise ValueError(
                    f"Cannot init node as user-owned node.\n"
                    f"Current API key does not permit action on user-owned nodes.\n"
                    f"Please provide a tag allowed by this API key: {authn_access_tags}"
                )
            access_blob_from_policy = {"user": identifier}
            access_blob_modified = True

        logger.info(
            f"Node to be initialized with access_blob: {access_blob_from_policy}"
        )
        # modified means the blob to-be-used was changed in comparison to the user input
        return access_blob_modified, access_blob_from_policy

    async def modify_node(
        self, node, principal, authn_access_tags, authn_scopes, access_blob
    ):
        if principal.type == "service":
            identifier = str(principal.uuid)
        else:
            identifier = self._get_id(principal)

        if access_blob == node.access_blob:
            logger.info(
                f"Node access_blob not modified; access_blob is identical: {access_blob}"
            )
            return False, node.access_blob

        if len(access_blob) != 1 or "tags" not in access_blob:
            raise ValueError(
                f"""access_blob must be in the form '{{"tags": ["tag1", "tag2", ...]}}'\n"""
                f"""Received {access_blob=}\n"""
                f"""If this was a merge patch on a user-owned node, use a replace op instead."""
            )
        if not access_blob["tags"]:
            if not self._is_admin(authn_scopes):
                raise ValueError(
                    "Cannot apply empty tag list to node: only Tiled admins can apply an empty tag list."
                )
        access_tags = set(access_blob["tags"])
        include_public_tag = False
        # check for tags that need to be added
        for tag in access_tags:
            if authn_access_tags is not None:
                if tag not in authn_access_tags:
                    raise ValueError(
                        f"Cannot apply tag to node: API key is restricted to access tags: {authn_access_tags}."
                    )
            if tag in node.access_blob.get("tags", []):
                # node already has this tag - no action.
                # or: access_blob does not have "tags" key,
                # so it must have a "user" key currently
                include_public_tag = include_public_tag or (
                    tag.casefold() == self.public_tag
                )
                continue
            elif tag.casefold() == self.public_tag:
                include_public_tag = True
                if not self._is_admin(authn_scopes):
                    raise ValueError(
                        "Cannot apply 'public' tag to node: only Tiled admins can apply the 'public' tag."
                    )
            elif not await self.is_tag_defined(tag):
                raise ValueError(f"Cannot apply tag to node: {tag=} is not defined")
            elif not await self.is_tag_owner(tag, identifier):
                # admins can ignore the tag ownership check
                if not self._is_admin(authn_scopes):
                    raise ValueError(
                        f"Cannot apply tag to node: user='{identifier}' is not an owner of {tag=}"
                    )
            elif tag.casefold() in self.invalid_tag_names:
                raise ValueError(
                    f"Cannot apply tag to node: '{tag}' is not a valid tag name."
                )

        access_tags_from_policy = {
            tag for tag in access_tags if tag.casefold() != self.public_tag
        }
        if include_public_tag:
            access_tags_from_policy.add(self.public_tag)

        # check for tags that need to be removed
        if "tags" in node.access_blob:
            for tag in set(node.access_blob["tags"]).difference(
                access_tags_from_policy
            ):
                if authn_access_tags is not None:
                    if tag not in authn_access_tags:
                        raise ValueError(
                            f"Cannot remove tag from node: "
                            f"API key is restricted to access tags: {authn_access_tags}."
                        )
                if tag == self.public_tag:
                    if not self._is_admin(authn_scopes):
                        raise ValueError(
                            "Cannot remove 'public' tag from node: only Tiled admins can remove the 'public' tag."
                        )
                elif not await self.is_tag_defined(tag):
                    raise ValueError(
                        f"Cannot remove tag from node: {tag=} is not defined"
                    )
                elif not await self.is_tag_owner(tag, identifier):
                    # admins can ignore the tag ownership check
                    if not self._is_admin(authn_scopes):
                        raise ValueError(
                            f"Cannot remove tag from node: user='{identifier}' is not an owner of {tag=}"
                        )
                elif tag.casefold() in self.invalid_tag_names:
                    raise ValueError(
                        f"Cannot remove tag from node: '{tag}' is not a valid tag name."
                    )

        access_blob_from_policy = {"tags": list(access_tags_from_policy)}
        access_blob_modified = access_tags != access_tags_from_policy

        # admin principals are not subject to scope reduction restriction
        if not self._is_admin(authn_scopes):
            # check that the access_blob change would not result in invalid scopes for user.
            # this applies when removing tags, but also must be done when
            # converting from user-owned node to shared (tagged) node
            new_scopes = set()
            for tag in access_tags_from_policy:
                new_scopes.update(await self.get_scopes_from_tag(tag, identifier))
            if not all(scope in new_scopes for scope in self.unremovable_scopes):
                raise ValueError(
                    f"Cannot modify tags on node: operation removes unremovable scopes.\n"
                    f"The current access_blob is: {node.access_blob}\n"
                    f"The new access_blob would be: {access_blob_from_policy}\n"
                    f"These scopes cannot be self-removed: {self.unremovable_scopes}"
                )

        logger.info(
            f"Node to be modified with new access_blob: {access_blob_from_policy}"
        )
        # modified means the blob to-be-used was changed in comparison to the user input
        return access_blob_modified, access_blob_from_policy

    async def allowed_scopes(self, node, principal, authn_access_tags, authn_scopes):
        # If this is being called, filter_for_access has let us get this far.
        # However, filters and allowed_scopes should always be implemented to
        # give answers consistent with each other.
        if not hasattr(node, "access_blob"):
            allowed = self.scopes
        elif self._is_admin(authn_scopes):
            allowed = self.scopes
        else:
            if principal is None:
                identifier = None
            elif principal.type == "service":
                identifier = str(principal.uuid)
            else:
                identifier = self._get_id(principal)

            allowed = set()
            if "user" in node.access_blob:
                if authn_access_tags is None and identifier == node.access_blob["user"]:
                    allowed = self.scopes
            elif "tags" in node.access_blob:
                for tag in node.access_blob["tags"]:
                    if authn_access_tags is not None:
                        if tag not in authn_access_tags:
                            continue
                    if await self.is_tag_public(tag):
                        allowed.update(self.read_scopes)
                        if tag == self.public_tag:
                            continue
                    elif not await self.is_tag_defined(tag):
                        continue
                    if identifier is not None:
                        tag_scopes = await self.get_scopes_from_tag(tag, identifier)
                        allowed.update(
                            tag_scopes if tag_scopes.issubset(self.scopes) else set()
                        )

        return allowed

    async def filters(self, node, principal, authn_access_tags, authn_scopes, scopes):
        queries = []
        query_filter = AccessBlobFilter

        if not hasattr(node, "access_blob"):
            return queries
        if not scopes.issubset(self.scopes):
            return NO_ACCESS

        tag_list = set()
        if principal is None:
            identifier = None
        else:
            if principal.type == "service":
                identifier = str(principal.uuid)
            elif self._is_admin(authn_scopes):
                return queries
            else:
                identifier = self._get_id(principal)
            tag_list.update(
                set.intersection(
                    *[
                        await self.get_tags_from_scope(scope, identifier)
                        for scope in scopes
                    ]
                )
            )

        tag_list.update(
            set.intersection(
                *[
                    await self.get_public_tags() if scope in self.read_scopes else set()
                    for scope in scopes
                ]
            )
        )

        if authn_access_tags is not None:
            identifier = None
            tag_list.intersection_update(authn_access_tags)

        queries.append(query_filter(identifier, tag_list))
        return queries



from logging import Filter, LogRecord

from ..utils import UNSET, SingleUserPrincipal
from .app import current_principal


class PrincipalFilter(Filter):
    """Logging filter to attach username or Service Principal UUID to LogRecord"""

    def filter(self, record: LogRecord) -> bool:
        principal = current_principal.get(UNSET)
        if principal is UNSET:
            # This will only occur if an uncaught exception was raised in the
            # server before the authentication code ran. This will always be
            # associated with a 500 Internal Server Error response.
            short_name = "unset"
        elif principal is None:
            short_name = "anon"
        elif principal is SingleUserPrincipal:
            short_name = "singleuser"
        elif principal.type == "service":
            short_name = f"service:{principal.uuid}"
        else:  # principal.type == "user"
            short_name = ",".join(
                f"'{identity.id}'" for identity in principal.identities
            )
        record.principal = short_name
        return True



from typing import Any, Optional

import dask.dataframe
import pandas

from ..catalog.orm import Node
from ..structures.data_source import DataSource
from .dataframe import DataFrameAdapter
from .mapping import MapAdapter


class ExcelAdapter(MapAdapter):
    @classmethod
    def from_file(cls, file: Any, **kwargs: Any) -> "ExcelAdapter":
        """
        Read the sheets in an Excel file.

        This maps the Excel file, which may contain one of more spreadsheets,
        onto a tree of tabular structures.

        Examples
        --------

        Given a file object

        >>> file = open("path/to/excel_file.xlsx")
        >>> ExcelAdapter.from_file(file)

        Given a pandas.ExcelFile object

        >>> import pandas
        >>> filepath = "path/to/excel_file.xlsx"
        >>> ef = pandas.ExcelFile(filepath)
        >>> ExcelAdapter.from_file(ef)

        Parameters
        ----------
        file :
        kwargs :

        Returns
        -------

        """
        if isinstance(file, pandas.ExcelFile):
            excel_file = file
        else:
            excel_file = pandas.ExcelFile(file)
        mapping = {}
        for sheet_name in excel_file.sheet_names:
            ddf = dask.dataframe.from_pandas(
                excel_file.parse(sheet_name),
                npartitions=1,  # TODO Be smarter about this.
            )
            mapping[sheet_name] = DataFrameAdapter.from_dask_dataframe(ddf)
        return cls(mapping, **kwargs)

    @classmethod
    def from_uris(cls, data_uri: str, **kwargs: Any) -> "ExcelAdapter":
        """
        Read the sheets in an Excel file.

        This maps the Excel file, which may contain one of more spreadsheets,
        onto a tree of tabular structures.

        Examples
        --------

        Given a file path

        >>> ExcelAdapter.from_file("path/to/excel_file.xlsx")

        Parameters
        ----------
        data_uri :
        kwargs :

        Returns
        -------

        """
        file = pandas.ExcelFile(data_uri)
        return cls.from_file(file)

    @classmethod
    def from_catalog(
        cls,
        # An Excel file is a container of tables, hence
        # DataSource[None].
        data_source: DataSource[None],
        node: Node,
        /,
        **kwargs: Optional[Any],
    ) -> "ExcelAdapter":
        data_uri = data_source.assets[0].data_uri
        return cls.from_uris(
            data_uri,
            structure=data_source.structure,
            metadata=node.metadata_,
            specs=node.specs,
            **kwargs,
        )



# Write the benchmarking functions here.
# See "Writing benchmarks" in the asv docs for more information.

import tempfile

import numpy
import pandas

from tiled.adapters.array import ArrayAdapter
from tiled.adapters.mapping import MapAdapter
from tiled.catalog import from_uri
from tiled.client import Context, from_context
from tiled.server.app import build_app
from tiled.structures.core import StructureFamily
from tiled.structures.data_source import DataSource
from tiled.structures.table import TableStructure


class TimeSuite:
    def setup(self):
        tree = MapAdapter({"x": ArrayAdapter.from_array(numpy.ones((100, 100)))})
        app = build_app(tree)
        self.context = Context.from_app(app)
        self.client = from_context(self.context)

    def teardown(self):
        self.context.close()

    def time_list_tree(self):
        list(self.client)

    def time_lookup(self):
        self.client["x"]

    def time_lookup_and_read(self):
        self.client["x"].read()


class CatalogSuite:
    def setup(self):
        self.directory = tempfile.TemporaryDirectory()
        self.df = pandas.DataFrame([])

        catalog = from_uri(
            f"sqlite:///{self.directory.name}/catalog.db",
            init_if_not_exists=True,
            writable_storage=self.directory.name,
        )
        self.context = Context.from_app(build_app(catalog))
        self.client = from_context(self.context)

    def teardown(self):
        self.context.close()
        self.directory.cleanup()

    def time_repeated_write(self):
        for _ in range(100):
            self.client.new(
                structure_family=StructureFamily.table,
                data_sources=[
                    DataSource(
                        structure_family=StructureFamily.table,
                        structure=TableStructure.from_pandas(self.df),
                        mimetype="text/csv",
                    ),  # or PARQUET_MIMETYPE
                ],
            )



import builtins
import math
import warnings
from abc import abstractmethod
from typing import Any, Iterable, List, Optional, Tuple, Union

import numpy as np
from ndindex import ndindex
from numpy._typing import NDArray

from ..catalog.orm import Node
from ..ndslice import NDSlice
from ..structures.array import ArrayStructure, BuiltinDtype
from ..structures.core import Spec, StructureFamily
from ..structures.data_source import DataSource
from ..type_aliases import JSON, EllipsisType
from ..utils import path_from_uri
from .utils import init_adapter_from_catalog


def force_reshape(arr: np.array, desired_shape: Tuple[int, ...]) -> np.array:
    """Reshape a numpy array to match the desired shape, if possible.

    Parameters
    ----------

    arr : np.array
        The original ND array to be reshaped
    desired_shape : Tuple[int, ...]
        The desired shape of the resulting array

    Returns
    -------

    A view of the original array
    """

    if arr.shape == desired_shape:
        # Nothing to do here
        return arr

    if arr.size == math.prod(desired_shape):
        if len(arr.shape) != len(desired_shape):
            # Missing or extra singleton dimensions
            warnings.warn(
                f"Forcefully reshaping {arr.shape} to {desired_shape}",
                category=RuntimeWarning,
            )
            return arr.reshape(desired_shape)
        else:
            # Some dimensions might be swapped or completely wrong
            # TODO: needs to be treated more carefully
            pass

    warnings.warn(
        f"Can not reshape array of {arr.shape} to match {desired_shape}; proceeding without changes",
        category=RuntimeWarning,
    )
    return arr


class FileSequenceAdapter:
    """Base adapter class for image (and other file) sequences

    Assumes that each file contains an array of the same shape and dtype, and the sequence of files defines the
    left-most dimension in the resulting compound (stacked) array.

    When subclassing, define the `_load_from_files` method specific for a particular file type.
    """

    structure_family = StructureFamily.array

    def __init__(
        self,
        data_uris: Iterable[str],
        *,
        structure: Optional[ArrayStructure] = None,
        metadata: Optional[JSON] = None,
        specs: Optional[List[Spec]] = None,
    ) -> None:
        """

        Parameters
        ----------
        seq :
        structure :
        metadata :
        specs :
        """
        self.filepaths = [path_from_uri(data_uri) for data_uri in data_uris]
        # TODO Check shape, chunks against reality.
        self.specs = specs or []
        self._provided_metadata = metadata or {}
        if structure is None:
            dat0 = self._load_from_files(0)
            shape = (len(self.filepaths), *dat0.shape[1:])
            structure = ArrayStructure(
                shape=shape,
                # one chunk per underlying image file
                chunks=((1,) * shape[0], *[(i,) for i in shape[1:]]),
                # Assume all files have the same data type
                data_type=BuiltinDtype.from_numpy_dtype(dat0.dtype),
            )
        self._structure = structure

    @classmethod
    def from_uris(cls, *data_uris: str) -> "FileSequenceAdapter":
        return cls(data_uris=data_uris)

    @classmethod
    def from_catalog(
        cls,
        data_source: DataSource[ArrayStructure],
        node: Node,
        /,
        **kwargs: Optional[Any],
    ) -> "FileSequenceAdapter":
        return init_adapter_from_catalog(cls, data_source, node, **kwargs)  # type: ignore

    @abstractmethod
    def _load_from_files(
        self, slice: Union[builtins.slice, int] = slice(None)
    ) -> NDArray[Any]:
        """Load the array data from files

        Parameters
        ----------
        slice : slice
            an optional slice along the left-most dimension in the resulting array; effectively selects a subset of
            files to be loaded

        Returns
        -------
            A numpy or dask ND array with data from each file stacked along an additional (left-most) dimension.
        """

        pass

    def metadata(self) -> JSON:
        """

        Returns
        -------

        """
        # TODO How to deal with the many headers?
        return self._provided_metadata

    def read(
        self, slice: Union[NDSlice, EllipsisType, builtins.slice] = ...
    ) -> NDArray[Any]:
        """Return a numpy array

        Receives a sequence of values to select from a collection of data files
        that were saved in a folder. The input order is defined as: files -->
        vertical slice --> horizontal slice --> color slice --> ... read() can
        receive one value or one slice to select all the data from one file or
        a sequence of files; or it can receive a tuple (int or slice) to select
        a more specific sequence of pixels of a group of images, for example.

        Parameters
        ----------
        slice : NDSlice, optional
            Specification of slicing to be applied to the data array

        Returns
        -------
            Return a numpy array
        """
        if slice is Ellipsis:
            arr = self._load_from_files()
        elif isinstance(slice, int):
            # e.g. read(slice=0) -- return an entire image (drop 0th dimension of the stack)
            arr = np.squeeze(self._load_from_files(slice), 0)
        elif isinstance(slice, builtins.slice):
            # e.g. read(slice=(...)) -- return a slice along the image axis
            arr = self._load_from_files(slice)
        elif isinstance(slice, tuple):
            if len(slice) == 0:
                arr = self._load_from_files()
            elif len(slice) == 1:
                arr = self.read(slice=slice[0])
            else:
                left_axis, *the_rest = slice
                # Could be int or slice (i, ...) or (slice(...), ...); the_rest is converted to a list
                if isinstance(left_axis, int):
                    # e.g. read(slice=(0, ....))
                    arr = np.squeeze(self._load_from_files(left_axis), 0)
                elif left_axis is Ellipsis:
                    # Return all images
                    arr = self._load_from_files()
                    the_rest.insert(0, Ellipsis)  # Include any leading dimensions
                elif isinstance(left_axis, builtins.slice):
                    arr = self.read(slice=left_axis)

                sliced_shape = ndindex(left_axis).newshape(self.structure().shape)
                arr = force_reshape(arr, sliced_shape)
                arr = np.atleast_1d(arr[tuple(the_rest)])
        else:
            raise RuntimeError(f"Unsupported slice type, {type(slice)} in {slice}")
        sliced_shape = ndindex(slice).newshape(self.structure().shape)
        return force_reshape(arr, sliced_shape)

    def read_block(
        self, block: Tuple[int, ...], slice: NDSlice = NDSlice(...)
    ) -> NDArray[Any]:
        """

        Parameters
        ----------
        block :
        slice :

        Returns
        -------

        """
        if any(block[1:]):
            raise IndexError(block)
        arr = self.read(builtins.slice(block[0], block[0] + 1))
        return arr[slice] if slice else arr

    def structure(self) -> ArrayStructure:
        """

        Returns
        -------

        """
        return self._structure



import awkward
import numpy
import pandas
import pytest
import sparse
import tifffile as tf
import xarray
from starlette.status import HTTP_409_CONFLICT

from ..catalog import in_memory
from ..client import Context, from_context
from ..client.composite import CompositeClient
from ..client.container import Container
from ..client.utils import ClientError
from ..server.app import build_app
from ..structures.array import ArrayStructure, BuiltinDtype
from ..structures.core import Spec, StructureFamily
from ..structures.data_source import Asset, DataSource, Management
from ..structures.table import TableStructure
from ..utils import ensure_uri
from .utils import fail_with_status_code

rng = numpy.random.default_rng(12345)

df1 = pandas.DataFrame({"A": ["one", "two", "three"], "B": [1, 2, 3]})
df2 = pandas.DataFrame(
    {
        "C": ["red", "green", "blue", "white"],
        "D": [10.0, 20.0, 30.0, 40.0],
        "E": [0, 0, 0, 0],
    }
)
df3 = pandas.DataFrame(
    {
        "col1": ["one", "two", "three", "four", "five"],
        "col2": [1.0, 2.0, 3.0, 4.0, 5.0],
    }
)
arr1 = rng.random(size=(3, 5), dtype="float64")
arr2 = rng.integers(0, 255, size=(5, 7, 3), dtype="uint8")
img_data = rng.integers(0, 255, size=(5, 13, 17, 3), dtype="uint8")

# An awkward array
awk_arr = awkward.Array(
    [
        [{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}],
        [],
        [{"x": 3.3, "y": [1, 2, 3]}],
    ]
)
awk_packed = awkward.to_packed(awk_arr)
awk_form, awk_length, awk_container = awkward.to_buffers(awk_packed)

# A sparse array
arr = rng.random(size=(10, 20, 30), dtype="float64")
sps_arr = sparse.COO(numpy.where(arr > 0.95, arr, 0))

md = {"md_key1": "md_val1", "md_key2": 2}


@pytest.fixture(scope="module")
def tree(tmp_path_factory):
    tempdir = tmp_path_factory.getbasetemp()
    return in_memory(
        writable_storage=[
            ensure_uri(tempdir / "data"),
            f"duckdb:///{tempdir / 'data.duckdb'}",
        ],
        readable_storage=[ensure_uri(tempdir)],
    )


@pytest.fixture(scope="module")
def context(tree):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        x = client.create_container(key="x", metadata=md, specs=["composite"])
        x.write_array(arr1, key="arr1", metadata={"md_key": "md_for_arr1"})
        x.write_array(arr2, key="arr2", metadata={"md_key": "md_for_arr2"})
        x.write_dataframe(
            df1,
            key="df1",
            metadata={
                "md_key": "md_for_df1",
                "A": {"md_key": "md_for_A"},
                "B": {"md_key": "md_for_B"},
            },
        )
        x.write_dataframe(
            df2,
            key="df2",
            metadata={
                "md_key": "md_for_df2",
                "C": {"md_key": "md_for_C"},
                "D": {"md_key": "md_for_D"},
                "E": {"md_key": "md_for_E"},
            },
        )
        x.write_awkward(awk_arr, key="awk", metadata={"md_key": "md_for_awk"})
        x.write_sparse(
            coords=sps_arr.coords,
            data=sps_arr.data,
            shape=sps_arr.shape,
            key="sps",
            metadata={"md_key": "md_for_sps"},
        )

        yield context


@pytest.fixture(scope="function")
def context_for_read(context):
    client = from_context(context)

    # Awkward arrays are not supported when building xarray, in general
    client["x"].delete_contents("awk", external_only=False)
    # Add an image array and a table with 5 rows
    client["x"].write_array(img_data, key="img")
    client["x"].write_dataframe(df3, key="df3")

    yield context

    # Restore the original context
    client["x"].write_awkward(awk_arr, key="awk", metadata={"md_key": "md_for_awk"})
    client["x"].delete_contents("img", external_only=False)
    client["x"].delete_contents("df3", external_only=False)


@pytest.fixture
def tiff_sequence(tmp_path_factory):
    tempdir = tmp_path_factory.mktemp("sequence-")
    filepaths = []
    for i in range(img_data.shape[0]):
        fpath = tempdir / f"temp{i:05}.tif"
        tf.imwrite(fpath, img_data[i, ...])
        filepaths.append(fpath)

    yield filepaths


@pytest.fixture
def csv_file(tmp_path_factory):
    tempdir = tmp_path_factory.mktemp("csv-")
    fpath = tempdir / "test.csv"
    df3.to_csv(fpath, index=False)

    yield fpath


@pytest.fixture
def tiff_data_source(tiff_sequence):
    tiff_assets = [
        Asset(
            data_uri=ensure_uri(fpath),
            is_directory=False,
            parameter="data_uris",
            num=i + 1,
        )
        for i, fpath in enumerate(tiff_sequence)
    ]
    tiff_structure_0 = ArrayStructure(
        data_type=BuiltinDtype.from_numpy_dtype(numpy.dtype("uint8")),
        shape=(5, 13, 17, 3),
        chunks=((1, 1, 1, 1, 1), (13,), (17,), (3,)),
    )
    yield DataSource(
        mimetype="multipart/related;type=image/tiff",
        assets=tiff_assets,
        structure_family=StructureFamily.array,
        structure=tiff_structure_0,
        management=Management.external,
    )


@pytest.fixture
def csv_data_source(csv_file):
    csv_assets = [
        Asset(
            data_uri=ensure_uri(csv_file),
            is_directory=False,
            parameter="data_uris",
        )
    ]
    yield DataSource(
        mimetype="text/csv",
        assets=csv_assets,
        structure_family=StructureFamily.table,
        structure=TableStructure.from_pandas(df3),
        management=Management.external,
    )


@pytest.mark.parametrize(
    "name, expected",
    [
        ("A", df1["A"]),
        ("B", df1["B"]),
        ("C", df2["C"]),
        ("D", df2["D"]),
        ("E", df2["E"]),
        ("arr1", arr1),
        ("arr2", arr2),
        ("awk", awk_arr),
        ("sps", sps_arr.todense()),
    ],
)
def test_reading(context, name, expected):
    client = from_context(context)
    actual = client["x"][name].read()
    if name == "sps":
        actual = actual.todense()
    assert numpy.array_equal(actual, expected)


def test_iterate_parts(context):
    client = from_context(context)
    for part in client["x"].base:
        client["x"].base[part].read()


def test_iterate_columns(context):
    client = from_context(context)
    for col, _client in client["x"].items():
        read_from_client = _client.read()
        read_from_column = client["x"][col].read()
        if col == "sps":
            read_from_client = read_from_client.todense()
            read_from_column = read_from_column.todense()
        assert numpy.array_equal(read_from_client, read_from_column)

        if col in ["arr1", "arr2", "awk", "sps"]:
            # Arrays can be read from the parent client
            assert client[f"x/{col}"].read() is not None
        else:
            # The column is not accessible from the parent client
            with pytest.raises(KeyError):
                client[f"x/{col}"].read()


def test_metadata(context):
    client = from_context(context)
    assert client["x"].metadata == md

    # Check metadata for each part
    for part in client["x"].base:
        c = client["x"].base[part]
        assert c.metadata["md_key"] == f"md_for_{part}"


def test_parts_not_directly_accessible(context):
    client = from_context(context)

    # The table is accessible as a part
    assert client["x"].base["df1"].read() is not None
    assert numpy.array_equal(client["x"].base["df1"]["A"].read(), df1["A"])
    assert numpy.array_equal(client["x"]["A"].read(), df1["A"])

    # The table is not accessible directly from the composite client
    with pytest.raises(KeyError):
        client["x"]["df1"].read()

    # The table is accessible from a parent client
    assert client["x/df1"].read() is not None
    assert client["x/df1/A"].read() is not None

    # The column is not accessible from the parent client
    with pytest.raises(KeyError):
        client["x/A"].read()


def test_external_assets(context, tiff_data_source, csv_data_source):
    client = from_context(context)

    y = client.create_container(key="y", specs=["composite"])
    y.new(
        structure_family=StructureFamily.array,
        data_sources=[tiff_data_source],
        key="image",
    )
    y.new(
        structure_family=StructureFamily.table,
        data_sources=[csv_data_source],
        key="table",
    )

    arr = y.base["image"].read()
    assert numpy.array_equal(arr, img_data)

    df = y.base["table"].read()
    for col in df.columns:
        assert numpy.array_equal(df[col], df3[col])

    assert set(y.keys()) == {"image", "col1", "col2"}


def test_read_full(context_for_read):
    client = from_context(context_for_read)
    ds = client["x"].read()

    assert isinstance(ds, xarray.Dataset)
    assert set(ds.data_vars) == {
        "arr1",
        "arr2",
        "A",
        "B",
        "C",
        "D",
        "E",
        "sps",
        "img",
        "col1",
        "col2",
    }


def test_read_selective(context_for_read):
    client = from_context(context_for_read)
    ds = client["x"].read(variables=["arr1", "arr2", "A", "B", "sps"])

    assert isinstance(ds, xarray.Dataset)
    assert set(ds.data_vars) == {"arr1", "arr2", "A", "B", "sps"}


@pytest.mark.parametrize("dim0", ["time", "col1"])
def test_read_selective_with_dim0(context_for_read, dim0):
    client = from_context(context_for_read)
    ds = client["x"].read(variables=["arr2", "img", "col1"], dim0=dim0)

    assert isinstance(ds, xarray.Dataset)
    assert set(ds.data_vars) == {"arr2", "img", "col1"}.difference([dim0])

    # Check the dimension names
    for var_name in ds.data_vars:
        assert ds[var_name].dims[0] == dim0


def test_delete_contents(context, tiff_data_source, csv_data_source):
    client = from_context(context)

    parts_before = len(client["x"].base)
    keys_before = len(client["x"].keys())
    assert parts_before == 6
    assert keys_before == 9

    # Attempt to delete an array that is internally managed
    with pytest.raises(ClientError):
        client["x"].delete_contents("arr1")

    # Delete a single part
    client["x"].delete_contents("arr1", external_only=False)
    assert "arr1" not in client["x"].base
    assert len(client["x"].base) == parts_before - 1
    assert len(client["x"].keys()) == keys_before - 1

    # Attempt to delete a DataFrame column
    assert "A" in client["x"].keys()
    with pytest.raises(KeyError):
        client["x"].delete_contents("A", external_only=False)
    client["x"].delete_contents("df1", external_only=False)
    assert "A" not in client["x"].keys()
    assert len(client["x"].base) == parts_before - 2
    assert len(client["x"].keys()) == keys_before - 3

    # Add and delete external data
    client["x"].new(
        structure_family=StructureFamily.array,
        data_sources=[tiff_data_source],
        key="image",
    )
    client["x"].new(
        structure_family=StructureFamily.table,
        data_sources=[csv_data_source],
        key="table",
    )
    # 6 original parts, 2 deleted, 2 added
    assert len(client["x"].base) == parts_before - 2 + 2
    assert len(client["x"].keys()) == keys_before - 3 + 3
    client["x"].delete_contents("image")
    assert "image" not in client["x"].base
    assert len(client["x"].base) == parts_before - 1
    assert len(client["x"].keys()) == keys_before - 1

    # Passing an empty list does not delete anything
    client["x"].delete_contents([])
    assert len(client["x"].base) == parts_before - 1
    assert len(client["x"].keys()) == keys_before - 1

    # Delete multiple parts
    client["x"].delete_contents(["arr2", "df2"], external_only=False)
    assert "arr2" not in client["x"].base
    assert "df1" not in client["x"].base

    # Delete all parts
    client["x"].delete_contents(external_only=False)
    assert len(client["x"].base) == 0
    assert len(client["x"].keys()) == 0


def test_write_one_table(tree):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        df = pandas.DataFrame({"A": [], "B": []})
        client.create_container(key="z", specs=["composite"])
        client["z"].write_dataframe(df)
        assert len(client["z"].base) == 1  # One table
        assert len(client["z"]) == 2  # Two columns


def test_write_two_tables(tree):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        df1 = pandas.DataFrame({"A": [], "B": []})
        df2 = pandas.DataFrame({"C": [], "D": [], "E": []})
        z = client.create_container(key="z", specs=["composite"])
        z.write_dataframe(df1, key="table1")
        z.write_dataframe(df2, key="table2")
        z.base["table1"].read()
        z.base["table2"].read()


def test_write_two_tables_colliding_names(tree):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        df1 = pandas.DataFrame({"A": [], "B": []})
        df2 = pandas.DataFrame({"C": [], "D": [], "E": []})
        z = client.create_container(key="z", specs=["composite"])
        z.write_dataframe(df1, key="table1")
        with fail_with_status_code(HTTP_409_CONFLICT):
            z.write_dataframe(df2, key="table1")


def test_write_two_tables_colliding_keys(tree):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        df1 = pandas.DataFrame({"A": [], "B": []})
        df2 = pandas.DataFrame({"A": [], "C": [], "D": []})
        z = client.create_container(key="z", specs=["composite"])
        z.write_dataframe(df1, key="table1")
        with pytest.raises(ValueError):
            z.write_dataframe(df2, key="table2")


def test_write_two_tables_two_arrays(tree):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        df1 = pandas.DataFrame({"A": [], "B": []})
        df2 = pandas.DataFrame({"C": [], "D": [], "E": []})
        arr1 = numpy.ones((5, 5), dtype=numpy.float64)
        arr2 = 2 * numpy.ones((5, 5), dtype=numpy.int8)
        z = client.create_container(key="z", specs=["composite"])

        # Write by data source.
        z.write_dataframe(df1, key="table1")
        z.write_dataframe(df2, key="table2")
        z.write_array(arr1, key="F")
        z.write_array(arr2, key="G")

        # Read by data source.
        z.base["table1"].read()
        z.base["table2"].read()
        z.base["F"].read()
        z.base["G"].read()

        # Read by column.
        for column in ["A", "B", "C", "D", "E", "F", "G"]:
            z[column].read()


def test_write_table_column_array_key_collision(tree):
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        df = pandas.DataFrame({"A": [], "B": []})
        arr = numpy.array([1, 2, 3], dtype=numpy.float64)

        z1 = client.create_container(key="z1", specs=["composite"])
        z1.write_dataframe(df, key="table1")
        with pytest.raises(ValueError):
            z1.write_array(arr, key="A")

        z2 = client.create_container(key="z2", specs=["composite"])
        z2.write_array(arr, key="A")
        with pytest.raises(ValueError):
            z2.write_dataframe(df, key="table1")


def test_composite_validator(tree):
    "Test the spec validator for marking existing containers"
    with Context.from_app(build_app(tree)) as context:
        client = from_context(context)
        y = client.create_container(key="y")  # No specs

        # 1. Assign spec to an empty container
        y.update_metadata(specs=["composite"])
        assert Spec("composite") in y.specs
        assert isinstance(client["y"], CompositeClient)
        # Revert the assignment
        y.update_metadata(specs=[])
        assert Spec("composite") not in y.specs
        assert isinstance(client["y"], Container)

        # 2. Created a nested container
        y.create_container(key="z")
        with pytest.raises(ClientError, match="Nested containers are not allowed"):
            y.update_metadata(specs=["composite"])
        y.delete_contents("z")

        # 3. Write some initial array data
        y.write_array(arr1, key="arr1")
        y.write_array(arr2, key="arr2")
        y.write_awkward(awk_arr, key="awk")
        y.write_sparse(
            coords=sps_arr.coords,
            data=sps_arr.data,
            shape=sps_arr.shape,
            key="sps",
        )

        # Composite spec can be assigned to a container with arrays
        y.update_metadata(specs=["composite"])
        assert Spec("composite") in y.specs
        assert isinstance(client["y"], CompositeClient)
        y.update_metadata(specs=[])

        # 4. Add two valid tables
        y.write_dataframe(df1, key="df1")
        y.write_dataframe(df2, key="df2")

        # Composite spec can be assigned to a container with arrays and tables
        y.update_metadata(specs=["composite"])
        assert Spec("composite") in y.specs
        assert isinstance(client["y"], CompositeClient)
        y.update_metadata(specs=[])

        # 5. Add an array with a conflicting name
        y.write_array(arr1, key="A")
        with pytest.raises(ClientError, match="Found conflicting names"):
            y.update_metadata(specs=["composite"])
        y.delete_contents("A", external_only=False)

        # 6. Add a table with a conflicting column names
        y.write_dataframe(df1, key="df1_copy")
        with pytest.raises(ClientError, match="Found conflicting names"):
            y.update_metadata(specs=["composite"])
        y.delete_contents("df1_copy", external_only=False)

        # 7. Composite spec cannot be used for tables or arrays
        err_message = "Composite spec can be assigned only to containers"
        for key in ["arr1", "awk", "sps", "df1"]:
            with pytest.raises(ClientError, match=err_message):
                y[key].update_metadata(specs=["composite"])
        with pytest.raises(ClientError, match=err_message):
            y.write_dataframe(df1, specs=["composite"])
        with pytest.raises(ClientError, match=err_message):
            y.write_array(arr1, specs=["composite"])
        with pytest.raises(ClientError, match=err_message):
            y.write_awkward(awk_arr, specs=["composite"])
        with pytest.raises(ClientError, match=err_message):
            y.write_sparse(
                coords=sps_arr.coords,
                data=sps_arr.data,
                shape=sps_arr.shape,
                specs=["composite"],
            )



from typing import List, Optional, Set

import pydantic_settings
from fastapi import HTTPException, Query, Request
from starlette.status import HTTP_403_FORBIDDEN, HTTP_404_NOT_FOUND, HTTP_410_GONE

from tiled.adapters.protocols import AnyAdapter
from tiled.server.schemas import Principal
from tiled.structures.core import StructureFamily

from ..type_aliases import Scopes
from ..utils import BrokenLink
from .core import NoEntry
from .utils import filter_for_access, record_timing


def get_root_tree(request: Request):
    return request.app.state.root_tree


async def get_entry(
    path: str,
    security_scopes: List[str],
    principal: Optional[Principal],
    authn_access_tags: Optional[Set[str]],
    authn_scopes: Scopes,
    root_tree: pydantic_settings.BaseSettings,
    session_state: dict,
    metrics: dict,
    structure_families: Optional[set[StructureFamily]] = None,
    access_policy=None,
) -> AnyAdapter:
    """
    Obtain a node in the tree from its path.

    Walk down the path starting from the root of the tree and filter
    access by the specified scopes.

    session_state is an optional dictionary passed in the session token
    """
    path_parts = [segment for segment in path.split("/") if segment]
    entry = root_tree
    # access_policy = getattr(request.app.state, "access_policy", None)
    # If the entry/adapter can take a session state, pass it in.
    # The entry/adapter may return itself or a different object.
    if hasattr(entry, "with_session_state") and session_state:
        entry = entry.with_session_state(session_state)
    # start at the root
    # filter and keep only what we are allowed to see from here
    entry = await filter_for_access(
        entry,
        access_policy,
        principal,
        authn_access_tags,
        authn_scopes,
        ["read:metadata"],
        metrics,
    )
    try:
        for i, segment in enumerate(path_parts):
            if hasattr(entry, "lookup_adapter"):
                # New catalog adapter
                # This adapter can jump directly to the node of interest,
                # but currenty doesn't, to ensure access_policy is applied.
                # Raises NoEntry or BrokenLink if the path is not found
                entry = await entry.lookup_adapter([segment])
            else:
                # Old-style dict-like interface
                # Traverse into sub-tree(s) to reach the desired entry
                try:
                    entry = entry[segment]
                except (KeyError, TypeError):
                    raise NoEntry(path_parts)

            # filter and keep only what we are allowed to see from here
            entry = await filter_for_access(
                entry,
                access_policy,
                principal,
                authn_access_tags,
                authn_scopes,
                ["read:metadata"],
                metrics,
            )

        # Now check that we have the requested scope according to the access policy
        if access_policy is not None:
            with record_timing(metrics, "acl"):
                allowed_scopes = await access_policy.allowed_scopes(
                    entry,
                    principal,
                    authn_access_tags,
                    authn_scopes,
                )
                if not set(security_scopes).issubset(allowed_scopes):
                    if "read:metadata" not in allowed_scopes:
                        # If you can't read metadata, it does not exist for you.
                        raise NoEntry(path_parts)
                    else:
                        # You can see this, but you cannot perform the requested
                        # operation on it.
                        raise HTTPException(
                            status_code=HTTP_403_FORBIDDEN,
                            detail=(
                                "Not enough permissions to perform this action on this node. "
                                f"Requires scopes {security_scopes}. "
                                f"Principal had scopes {list(allowed_scopes)} on this node."
                            ),
                        )
    except BrokenLink as err:
        raise HTTPException(status_code=HTTP_410_GONE, detail=err.args[0])
    except NoEntry:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND, detail=f"No such entry: {path_parts}"
        )
    # Fast path for the common successful case
    if (structure_families is None) or (entry.structure_family in structure_families):
        return entry
    raise HTTPException(
        status_code=HTTP_404_NOT_FOUND,
        detail=(
            f"The node at {path} has structure family {entry.structure_family} "
            "and this endpoint is compatible with structure families "
            f"{structure_families}"
        ),
    )


def block(
    # Ellipsis as the "default" tells FastAPI to make this parameter required.
    block: str = Query(..., pattern="^[0-9]*(,[0-9]+)*$"),
):
    "Specify and parse a block index parameter."
    if not block:
        return ()
    return tuple(map(int, block.split(",")))


def expected_shape(
    expected_shape: Optional[str] = Query(
        None, min_length=1, pattern="^[0-9]+(,[0-9]+)*$|^scalar$"
    ),
):
    "Specify and parse an expected_shape parameter."
    if expected_shape is None:
        return
    if expected_shape == "scalar":
        return ()
    return tuple(map(int, expected_shape.split(",")))


def shape_param(
    shape: str = Query(..., min_length=1, pattern="^[0-9]+(,[0-9]+)*$|^scalar$"),
):
    "Specify and parse a shape parameter."
    return tuple(map(int, shape.split(",")))


def offset_param(
    offset: str = Query(..., min_length=1, pattern="^[0-9]+(,[0-9]+)*$"),
):
    "Specify and parse an offset parameter."
    return tuple(map(int, offset.split(",")))



import getpass
import os
import re
import sys
import time
import urllib.parse
import warnings
from pathlib import Path
from typing import List, Literal
from urllib.parse import parse_qs, urlparse

import httpx
import platformdirs
from pydantic import TypeAdapter

from tiled.schemas import About, AboutAuthenticationProvider

from .._version import __version__ as tiled_version
from ..utils import UNSET, DictView, parse_time_string
from .auth import CannotRefreshAuthentication, TiledAuth, build_refresh_request
from .decoders import SUPPORTED_DECODERS
from .transport import Transport
from .utils import (
    DEFAULT_TIMEOUT_PARAMS,
    MSGPACK_MIME_TYPE,
    handle_error,
    retry_context,
)

USER_AGENT = f"python-tiled/{tiled_version}"
API_KEY_AUTH_HEADER_PATTERN = re.compile(r"^Apikey (\w+)$")


def raise_if_cannot_prompt():
    if not _can_prompt() and not bool(int(os.environ.get("TILED_FORCE_PROMPT", "0"))):
        raise CannotPrompt(
            """
Tiled has detected that it is running in a 'headless' context where it cannot
prompt the user to provide credentials in the stdin. Options:

- Provide an API key in the environment variable TILED_API_KEY for Tiled to
  use.

- If Tiled has detected wrongly, set the environment variable
  TILED_FORCE_PROMPT=1 to override and force an interactive prompt.

- If you are developing an application that is wraping Tiled,
  obtain tokens using functions tiled.client.context.password_grant
  and/or device_code_grant, and pass them like Context.authenticate(tokens=token).
"""
        )


def identity_provider_input(
    providers: List[AboutAuthenticationProvider],
) -> AboutAuthenticationProvider:
    while True:
        print("Authentication providers:")
        for i, spec in enumerate(providers, start=1):
            print(f"{i} - {spec.provider}")
        raw_choice = input(
            "Choose an authentication provider (or press Enter to cancel): "
        )
        if not raw_choice:
            print("No authentication provider chosen. Failed.")
            break
        try:
            choice = int(raw_choice)
        except TypeError:
            print("Choice must be a number.")
            continue
        try:
            spec = providers[choice - 1]
        except IndexError:
            print(f"Choice must be a number 1 through {len(providers)}.")
            continue
        break
    return spec


def username_input():
    raise_if_cannot_prompt()
    return input("Username: ")


def password_input():
    raise_if_cannot_prompt()
    return getpass.getpass()


class PasswordRejected(RuntimeError):
    pass


def prompt_for_credentials(http_client, providers: List[AboutAuthenticationProvider]):
    """
    Prompt for credentials or third-party login at an interactive terminal.
    """
    if len(providers) == 1:
        # There is only one choice, so no need to prompt the user.
        spec = providers[0]
    else:
        spec = identity_provider_input(providers)
    auth_endpoint = spec.links["auth_endpoint"]
    provider = spec.provider
    mode = spec.mode
    # Note: "password" is included here for back-compat with older servers;
    # the new name for this mode is "internal".
    if mode == "internal" or mode == "password":
        # Prompt for username, password at terminal.
        username = username_input()
        PASSWORD_ATTEMPTS = 3
        for _attempt in range(PASSWORD_ATTEMPTS):
            password = password_input()
            if not password:
                raise PasswordRejected("Password empty.")
            try:
                tokens = password_grant(
                    http_client, auth_endpoint, provider, username, password
                )
            except httpx.HTTPStatusError as err:
                if err.response.status_code == httpx.codes.UNAUTHORIZED:
                    print(
                        "Username or password not recognized. Retry, or press Enter to cancel."
                    )
                    continue
                raise
            else:
                # Success! We have tokens.
                break
        else:
            # All attempts failed.
            raise PasswordRejected
    elif mode == "external":
        # Display link and access code, and try to open web browser.
        # Block while polling the server awaiting confirmation of authorization.
        tokens = device_code_grant(http_client, auth_endpoint)
    else:
        raise ValueError(f"Server has unknown authentication mechanism {mode!r}")
    confirmation_message = spec.confirmation_message
    if confirmation_message:
        username = tokens["identity"]["id"]
        print(confirmation_message.format(id=username))
    return tokens


class Context:
    """
    Wrap an httpx.Client with an optional cache and authentication functionality.
    """

    def __init__(
        self,
        uri,
        *,
        headers=None,
        api_key=None,
        cache=UNSET,
        timeout=None,
        verify=True,
        app=None,
        raise_server_exceptions=True,
    ):
        # The uri is expected to reach the root API route.
        uri = httpx.URL(uri)
        headers = headers or {}
        # Define this here instead of at module scope so that the SUPPORTED_DECODERS
        # may be modified before Context instantiation.
        ACCEPT_ENCODING = ", ".join(
            [key for key in SUPPORTED_DECODERS.keys() if key != "identity"]
        )
        # Resolve this here, not at module scope, because the test suite
        # injects TILED_CACHE_DIR env var to use a temporary directory.
        TILED_CACHE_DIR = Path(
            os.getenv("TILED_CACHE_DIR", platformdirs.user_cache_dir("tiled"))
        )
        headers.setdefault("accept-encoding", ACCEPT_ENCODING)
        # Set the User Agent to help the server fail informatively if the client
        # version is too old.
        headers.setdefault("user-agent", USER_AGENT)

        # If ?api_key=... is present, move it from the query into a header.
        # The server would accept it in the query parameter, but using
        # a header is a little more secure (e.g. not logged).
        parsed_params = urllib.parse.parse_qs(uri.query.decode())
        api_key_list = parsed_params.pop("api_key", None)
        if api_key_list is not None:
            if api_key is not None:
                raise ValueError(
                    "api_key was provided as query parameter in URI and as keyword argument. Pick one."
                )
            if len(api_key_list) != 1:
                raise ValueError("Cannot handle two api_key query parameters")
            (api_key,) = api_key_list
        if api_key is None:
            # Check for an API key from the environment.
            api_key = os.getenv("TILED_API_KEY")
        # We will set the API key via the `api_key` property below,
        # after constructing the Client object.

        # FastAPI redirects /api -> /api/ so add it here to save a request.
        path = uri.path
        if not path.endswith("/"):
            path = f"{path}/"
        # Construct the uri *without* api_key param.
        # Drop any params/fragments.
        self.api_uri = httpx.URL(
            urllib.parse.urlunsplit((uri.scheme, uri.netloc.decode(), path, {}, ""))
        )
        if timeout is None:
            timeout = httpx.Timeout(**DEFAULT_TIMEOUT_PARAMS)
        if cache is UNSET:
            cache = None
        if app is None:
            client = httpx.Client(
                transport=Transport(cache=cache),
                verify=verify,
                timeout=timeout,
                follow_redirects=True,
            )
            # Do this in the setter to avoid being overwritten.
            client.headers = headers
        else:
            # Set up an ASGI client.
            # Because we have been handed an app, we can infer that
            # starlette is available.
            from starlette.testclient import TestClient

            base_uri = f"{uri.scheme}://{uri.netloc}"
            # verify parameter is dropped, as there is no SSL in ASGI mode
            client = TestClient(
                app=app,
                raise_server_exceptions=raise_server_exceptions,
                base_url=base_uri,
            )
            client.timeout = timeout
            client.headers = headers
            # Do this in the setter to avoid being overwritten.
            client.follow_redirects = True
            client._transport = Transport(transport=client._transport, cache=cache)
            client.__enter__()
            # The TestClient is meant to be used only as a context manager,
            # where the context starts and stops and the wrapped ASGI app.
            # We are abusing it slightly to enable interactive use of the
            # TestClient.

            import threading

            # The threading module has its own (internal) atexit
            # mechanism that runs at thread shutdown, prior to the atexit
            # mechanism that runs at interpreter shutdown.
            # We need to intervene at that layer to close the portal, or else
            # we will wait forever for a thread run by the portal to join().
            threading._register_atexit(client.__exit__)

        self.http_client = client
        self._verify = verify
        self._cache = cache
        self._token_cache = Path(TILED_CACHE_DIR / "tokens")

        # Make an initial "safe" request to:
        # (1) Get the server_info.
        # (2) Let the server set the CSRF cookie.
        # No authentication has been set up yet, so these requests will be unauthenticated.
        # https://cheatsheetseries.owasp.org/cheatsheets/Cross-Site_Request_Forgery_Prevention_Cheat_Sheet.html#double-submit-cookie
        for attempt in retry_context():
            with attempt:
                server_info = handle_error(
                    self.http_client.get(
                        self.api_uri,
                        headers={
                            "Accept": MSGPACK_MIME_TYPE,
                            "Cache-Control": "no-cache, no-store",
                        },
                    )
                ).json()
        self.server_info: About = TypeAdapter(About).validate_python(server_info)
        self.api_key = api_key  # property setter sets Authorization header
        self.admin = Admin(self)  # accessor for admin-related requests

    def __repr__(self):
        auth_info = []
        if (self.api_key is None) and (self.http_client.auth is None):
            auth_info.append("(unauthenticated)")
        else:
            auth_info.append("authenticated")
            if self.server_info.authentication.links:
                whoami = self.whoami()
                auth_info.append("as")
                if whoami["type"] == "service":
                    auth_info.append(f"service '{whoami['uuid']}'")
                else:
                    auth_info.append(
                        ",".join(
                            f"'{identity['id']}'" for identity in whoami["identities"]
                        )
                    )
            if self.api_key is not None:
                auth_info.append(
                    f"with API key '{self.api_key[:min(len(self.api_key)//2, 8)]}...'"
                )
        auth_repr = " ".join(auth_info)
        return f"<{type(self).__name__} {auth_repr}>"

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        self.http_client.__exit__()

    def __getstate__(self):
        if getattr(self.http_client, "app", None):
            raise TypeError(
                "Cannot pickle a Tiled Context built around an ASGI. "
                "Only Tiled Context connected to remote servers can be pickled."
            )
        return (
            tiled_version,
            self.api_uri,
            self.http_client.headers,
            list(self.http_client.cookies.jar),
            self.http_client.timeout,
            self.http_client.auth,
            self._verify,
            self._token_cache,
            self.server_info,
            self.cache,
        )

    def __setstate__(self, state):
        (
            state_tiled_version,
            api_uri,
            headers,
            cookies_list,
            timeout,
            auth,
            verify,
            token_cache,
            server_info,
            cache,
        ) = state
        if state_tiled_version != tiled_version:
            raise TypeError(
                f"Cannot unpickle {type(self).__name__} from tiled version {state_tiled_version} "
                f"using tiled version {tiled_version}. Pickle should only be used to short-term "
                "transfer between identical versions of tiled."
            )
        self.api_uri = api_uri
        cookies = httpx.Cookies()
        for cookie in cookies_list:
            cookies.set(
                cookie.name, cookie.value, domain=cookie.domain, path=cookie.path
            )
        self.http_client = httpx.Client(
            verify=verify,
            transport=Transport(cache=cache),
            cookies=cookies,
            timeout=timeout,
            headers=headers,
            follow_redirects=True,
            auth=auth,
        )
        self._token_cache = token_cache
        self._cache = cache
        self._verify = verify
        self.server_info = server_info

    @classmethod
    def from_any_uri(
        cls,
        uri,
        *,
        headers=None,
        api_key=None,
        cache=UNSET,
        timeout=None,
        verify=True,
        app=None,
    ):
        """
        Accept a URI to a specific node.

        For example, given URI "https://example.com/api/v1//metadata/a/b/c"
        return a Context connected to "https://examples/api/v1" and the list
        ["a", "b", "c"].
        """
        uri = httpx.URL(uri)
        node_path_parts = []
        # Ensure that HTTPS is used if available
        # Logic will follow only one redirect, it is intended ONLY to toggle HTTPS.
        # The redirect will be followed only if the netloc host is identical to the original.
        if uri.scheme == "http":
            for attempt in retry_context():
                with attempt:
                    response_from_http = httpx.get(uri)
            if response_from_http.is_redirect:
                redirect_uri = httpx.URL(response_from_http.headers["location"])
                if redirect_uri.scheme == "https" and redirect_uri.host == uri.host:
                    uri = redirect_uri
        if "/metadata" in uri.path:
            api_path, _, node_path = uri.path.partition("/metadata")
            api_uri = uri.copy_with(path=api_path)
            node_path_parts.extend(
                [segment for segment in node_path.split("/") if segment]
            )
        # Below we handle the case where we are given *less* than the full URI
        # to the root endpoint. Here we are taking some care to plan for the case
        # where tiled is served at a sub-path, even though that is not yet supported
        # on the server side.
        elif "/api" not in uri.path:
            # Looks like we were given the root path (to the HTML landing page).
            path = uri.path
            if path.endswith("/"):
                path = path[:-1]
            api_uri = uri.copy_with(path=f"{path}/api/v1")
        elif "/v1" not in uri.path:
            # Looks like we were given the /api but no version.
            path = uri.path
            if path.endswith("/"):
                path = path[:-1]
            api_uri = uri.copy_with(path=f"{path}/v1")
        else:
            api_uri = uri
        context = cls(
            api_uri,
            headers=headers,
            api_key=api_key,
            cache=cache,
            timeout=timeout,
            verify=verify,
            app=app,
        )
        return context, node_path_parts

    @classmethod
    def from_app(
        cls,
        app,
        *,
        cache=UNSET,
        headers=None,
        timeout=None,
        api_key=UNSET,
        raise_server_exceptions=True,
        uri=None,
    ):
        """
        Construct a Context around a FastAPI app. Primarily for testing.
        """
        context = cls(
            uri="http://local-tiled-app/api/v1" if not uri else uri,
            headers=headers,
            api_key=None,
            cache=cache,
            timeout=timeout,
            app=app,
            raise_server_exceptions=raise_server_exceptions,
        )
        if api_key is UNSET:
            if not context.server_info.authentication.providers:
                # This is a single-user server.
                # Extract the API key from the app and set it.
                from ..server.settings import get_settings

                settings = app.dependency_overrides[get_settings]()
                api_key = settings.single_user_api_key or None
            else:
                # This is a multi-user server but no API key was passed,
                # so we will leave it as None on the Context.
                api_key = None
        context.api_key = api_key
        return context

    @property
    def tokens(self) -> DictView[Literal["access_token", "refresh_token"], str]:
        "A view of the current access and refresh tokens."
        return DictView(self.http_client.auth.tokens)

    @property
    def api_key(self):
        # Extract from header to ensure that there is one "ground truth" here
        # and no possibility of state getting out of sync.
        header = self.http_client.headers.get("Authorization", "")
        match = API_KEY_AUTH_HEADER_PATTERN.match(header)
        if match is not None:
            return match.group(1)

    @api_key.setter
    def api_key(self, api_key):
        if api_key is None:
            if self.http_client.headers.get("Authorization", "").startswith("Apikey "):
                self.http_client.headers.pop("Authorization")
        else:
            self.http_client.headers["Authorization"] = f"Apikey {api_key}"

    @property
    def cache(self):
        return self.http_client._transport.cache

    @cache.setter
    def cache(self, cache):
        self.http_client._transport.cache = cache

    def which_api_key(self):
        """
        A 'who am I' for API keys
        """
        if not self.api_key:
            raise RuntimeError("Not API key is configured for the client.")
        for attempt in retry_context():
            with attempt:
                return handle_error(
                    self.http_client.get(
                        self.server_info.authentication.links.apikey,
                        headers={"Accept": MSGPACK_MIME_TYPE},
                    )
                ).json()

    def create_api_key(self, scopes=None, expires_in=None, note=None, access_tags=None):
        """
        Generate a new API key.

        Users with administrative scopes may use ``Context.admin.revoke_api_key``
        to create API keys on behalf of other users or services.

        Parameters
        ----------
        scopes : Optional[List[str]]
            Restrict the access available to the API key by listing specific scopes.
            By default, this will have the same access as the user.
        expires_in : Optional[Union[int, str]]
            Number of seconds until API key expires, given as integer seconds
            or a string like: '3y' (years), '3d' (days), '5m' (minutes), '1h'
            (hours), '30s' (seconds). If None, it will never expire or it will
            have the maximum lifetime allowed by the server.
        note : Optional[str]
            Description (for humans).
        access_tags : Optional[List[str]]
            Restrict the access available to the API key by listing specific tags.
            By default, this will have no limits on access tags.
        """
        if isinstance(expires_in, str):
            expires_in = parse_time_string(expires_in)
        for attempt in retry_context():
            with attempt:
                return handle_error(
                    self.http_client.post(
                        self.server_info.authentication.links.apikey,
                        headers={"Accept": MSGPACK_MIME_TYPE},
                        json={
                            "scopes": scopes,
                            "access_tags": access_tags,
                            "expires_in": expires_in,
                            "note": note,
                        },
                    )
                ).json()

    def revoke_api_key(self, first_eight):
        """
        Revoke an API key.

        The API key must belong to the currently-authenticated user or service.
        Users with administrative scopes may use ``Context.admin.revoke_api_key``
        to revoke API keys belonging to other users.

        Parameters
        ----------
        first_eight : str
            Identify the API key to be deleted by passing its first 8 characters.
            (Any additional characters passed will be truncated.)
        """
        url_path = self.server_info.authentication.links.apikey
        for attempt in retry_context():
            with attempt:
                handle_error(
                    self.http_client.delete(
                        url_path,
                        headers={"x-csrf": self.http_client.cookies["tiled_csrf"]},
                        params={
                            **parse_qs(urlparse(url_path).query),
                            "first_eight": first_eight[:8],
                        },
                    )
                )

    @property
    def app(self):
        warnings.warn(
            "The 'app' accessor on Context is deprecated. Use Context.http_client.app.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.http_client.app

    @property
    def base_url(self):
        warnings.warn(
            "The 'base_url' accessor on Context is deprecated. Use Context.http_client.base_url.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.http_client.base_url

    @property
    def event_hooks(self):
        "httpx.Client event hooks. This is exposed for testing."
        warnings.warn(
            "The 'event_hooks' accessor on Context is deprecated. Use Context.http_client.event_hooks.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.http_client.event_hooks

    def authenticate(
        self,
        *,
        remember_me=True,
    ):
        """
        Log in to a Tiled server.

        Depending on the server's authentication method, this will prompt for username/password:

        >>> c.login()
        Username: USERNAME
        Password: <input is hidden>

        or prompt you to open a link in a web browser to login with a third party:

        >>> c.login()
        You have ... minutes to visit this URL

        https://...

        and enter the code: XXXX-XXXX

        Parameters
        ----------
        remember_me : bool
            Next time, try to automatically authenticate using this session.
        """
        # Obtain tokens via OAuth2 unless the caller has passed them.
        providers = self.server_info.authentication.providers
        tokens = prompt_for_credentials(
            self.http_client,
            providers,
        )
        self.configure_auth(tokens, remember_me=remember_me)

    # These two methods are aliased for convenience.
    login = authenticate

    def configure_auth(self, tokens, remember_me=True):
        """
        Configure Tiled client with tokens for refresh flow.

        Parameters
        ----------
        tokens : dict, optional
            Must include keys 'access_token' and 'refresh_token'
        """
        self.http_client.auth = None
        if self.api_key is not None:
            raise RuntimeError(
                "An API key is set. Cannot use both API key and OAuth2 authentication."
            )
        # Configure an httpx.Auth instance on the http_client, which
        # will manage refreshing the tokens as needed.
        refresh_url = self.server_info.authentication.links.refresh_session
        csrf_token = self.http_client.cookies["tiled_csrf"]
        if remember_me:
            token_directory = self._token_directory()
        else:
            # Clear any existing tokens.
            temp_auth = TiledAuth(refresh_url, csrf_token, self._token_directory())
            temp_auth.sync_clear_token("access_token")
            temp_auth.sync_clear_token("refresh_token")
            # Store tokens in memory only, with no syncing to disk.
            token_directory = None
        auth = TiledAuth(refresh_url, csrf_token, token_directory)
        auth.sync_set_token("access_token", tokens["access_token"])
        auth.sync_set_token("refresh_token", tokens["refresh_token"])
        self.http_client.auth = auth

    @property
    def authenticated(self) -> bool:
        """
        Boolean indicated whether session is authenticated (true) or anonymous (false)

        Examples
        --------

        An anonymous session at first, after login, is authenticated.

        >>> client.context.authenticated
        False
        >>> client.login()
        Username: USERNAME
        Password: <input is hidden>
        >>> client.context.authenticated
        True

        """
        # Confirm the state of properties that authentication consists of
        return (self.api_key is not None) or (self.http_client.auth is not None)

    def _token_directory(self):
        # e.g. ~/.config/tiled/tokens/{host:port}
        # with the templated element URL-encoded so it is a valid filename.
        path = Path(
            self._token_cache,
            urllib.parse.quote_plus(str(self.api_uri)),
        )

        # If this directory already exists, it might contain subdirectories
        # left by older versions of tiled that supported caching tokens for
        # multiple users of one server. Clean them up.
        if path.is_dir():
            import shutil

            [shutil.rmtree(item) for item in path.iterdir() if item.is_dir()]

        return path

    def use_cached_tokens(self):
        """
        Attempt to reconnect using cached tokens.

        Returns
        -------
        success : bool
            Indicating whether valid cached tokens were found
        """
        refresh_url = self.server_info.authentication.links.refresh_session
        csrf_token = self.http_client.cookies["tiled_csrf"]

        # Try automatically authenticating using cached tokens, if any.
        token_directory = self._token_directory()
        # We have to make an HTTP request to let the server validate whether we
        # have a valid session.
        self.http_client.auth = TiledAuth(refresh_url, csrf_token, token_directory)
        # This will either:
        # * Use an access_token and succeed.
        # * Use a refresh_token to attempt refresh flow and succeed.
        # * Use a refresh_token to attempt refresh flow and fail, raise.
        # * Find no tokens and raise.
        try:
            self.whoami()
            return True
        except CannotRefreshAuthentication:
            self.http_client.auth = None
            return False

    def force_auth_refresh(self):
        """
        Execute refresh flow.

        This method is exposed for testing and debugging uses.

        It should never be necessary for the user to call. Refresh flow is
        automatically executed by tiled.client.auth.TiledAuth when the current
        access_token expires.
        """
        if self.http_client.auth is None:
            raise RuntimeError(
                "No authentication has been set up. Cannot reauthenticate."
            )
        refresh_token = self.http_client.auth.sync_get_token(
            "refresh_token", reload_from_disk=True
        )
        if refresh_token is None:
            raise CannotRefreshAuthentication("There is no refresh_token.")
        csrf_token = self.http_client.cookies["tiled_csrf"]
        refresh_request = build_refresh_request(
            self.http_client.auth.refresh_url,
            refresh_token,
            csrf_token,
        )
        for attempt in retry_context():
            with attempt:
                token_response = self.http_client.send(refresh_request, auth=None)
                if token_response.status_code == httpx.codes.UNAUTHORIZED:
                    raise CannotRefreshAuthentication(
                        "Session cannot be refreshed. Log in again."
                    )
                handle_error(token_response)
        tokens = token_response.json()
        self.http_client.auth.sync_set_token("access_token", tokens["access_token"])
        self.http_client.auth.sync_set_token("refresh_token", tokens["refresh_token"])
        return tokens

    def whoami(self):
        "Return information about the currently-authenticated user or service."
        for attempt in retry_context():
            with attempt:
                return handle_error(
                    self.http_client.get(
                        self.server_info.authentication.links.whoami,
                        headers={"Accept": MSGPACK_MIME_TYPE},
                    )
                ).json()

    def logout(self):
        """
        Log out of the current session (if any).

        This method is idempotent.
        """
        if self.http_client.auth is None:
            return

        # Revoke the current session.
        refresh_token = self.http_client.auth.sync_get_token("refresh_token")
        for attempt in retry_context():
            with attempt:
                handle_error(
                    self.http_client.post(
                        f"{self.api_uri}auth/session/revoke",
                        json={"refresh_token": refresh_token},
                        # Circumvent auth because this request is not authenticated.
                        # The refresh_token in the body is the relevant proof, not the
                        # 'Authentication' header.
                        auth=None,
                    )
                )

        # Clear on-disk state.
        self.http_client.auth.sync_clear_token("access_token")
        self.http_client.auth.sync_clear_token("refresh_token")

        # Clear in-memory state.
        self.http_client.headers.pop("Authorization", None)
        self.http_client.auth = None

    def revoke_session(self, session_id):
        """
        Revoke a Session so it cannot be refreshed.

        This may be done to ensure that a possibly-leaked refresh token cannot be used.
        """
        for attempt in retry_context():
            with attempt:
                handle_error(
                    self.http_client.delete(
                        self.server_info.authentication.links.revoke_session.format(
                            session_id=session_id
                        ),
                        headers={"x-csrf": self.http_client.cookies["tiled_csrf"]},
                    )
                )


class Admin:
    "Accessor for requests that require administrative privileges."

    def __init__(self, context: Context):
        self.context = context
        self.base_url = context.server_info.links["self"]

    def list_principals(self, offset=0, limit=100):
        "List Principals (users and services) in the authentication database."
        url_path = f"{self.base_url}/auth/principal"
        params = {
            **parse_qs(urlparse(url_path).query),
            "offset": offset,
            "limit": limit,
        }
        for attempt in retry_context():
            with attempt:
                return handle_error(
                    self.context.http_client.get(url_path, params=params)
                ).json()

    def show_principal(self, uuid):
        "Show one Principal (user or service) in the authentication database."
        for attempt in retry_context():
            with attempt:
                return handle_error(
                    self.context.http_client.get(
                        f"{self.base_url}/auth/principal/{uuid}"
                    )
                ).json()

    def create_api_key(
        self, uuid, scopes=None, expires_in=None, note=None, access_tags=None
    ):
        """
        Generate a new API key for another user or service.

        Parameters
        ----------
        uuid : str
            Identify the principal -- the user or service
        scopes : Optional[List[str]]
            Restrict the access available to the API key by listing specific scopes.
            By default, this will have the same access as the principal.
        expires_in : Optional[int]
            Number of seconds until API key expires. If None,
            it will never expire or it will have the maximum lifetime
            allowed by the server.
        note : Optional[str]
            Description (for humans).
        access_tags : Optional[List[str]]
            Restrict the access available to the API key by listing specific tags.
            By default, this will have no limits on access tags.
        """
        for attempt in retry_context():
            with attempt:
                return handle_error(
                    self.context.http_client.post(
                        f"{self.base_url}/auth/principal/{uuid}/apikey",
                        headers={"Accept": MSGPACK_MIME_TYPE},
                        json={
                            "scopes": scopes,
                            "access_tags": access_tags,
                            "expires_in": expires_in,
                            "note": note,
                        },
                    )
                ).json()

    def create_service_principal(
        self,
        role,
    ):
        """
        Generate a new service principal.

        Parameters
        ----------
        role : str
            Specify the role (e.g. user or admin)
        """
        url_path = f"{self.base_url}/auth/principal"
        for attempt in retry_context():
            with attempt:
                return handle_error(
                    self.context.http_client.post(
                        url_path,
                        headers={"Accept": MSGPACK_MIME_TYPE},
                        params={**parse_qs(urlparse(url_path).query), "role": role},
                    )
                ).json()

    def revoke_api_key(self, uuid, first_eight=None):
        """
        Revoke an API key belonging to any user or service.

        Parameters
        ----------
        uuid : str
            Identify the principal whose API key will be deleted. This is
            required in order to reduce the chance of accidentally revoking
            the wrong key.
        first_eight : str
            Identify the API key to be deleted by passing its first 8 characters.
            (Any additional characters passed will be truncated.)
        """
        url_path = f"{self.base_url}/auth/principal/{uuid}/apikey"
        for attempt in retry_context():
            with attempt:
                return handle_error(
                    self.context.http_client.delete(
                        url_path,
                        headers={"Accept": MSGPACK_MIME_TYPE},
                        params={
                            **parse_qs(urlparse(url_path).query),
                            "first_eight": first_eight[:8],
                        },
                    )
                )


class CannotPrompt(Exception):
    pass


def _can_prompt():
    "Infer whether the user can be prompted for a password or user code."

    if (not sys.__stdin__.closed) and sys.__stdin__.isatty():
        return True
    # In IPython (TerminalInteractiveShell) the above is true, but in
    # Jupyter (ZMQInteractiveShell) it is False.
    # Jupyter own mechanism for giving a prompt, so we always return
    # True if we detect IPython/Jupyter.
    if "IPython" in sys.modules:
        import IPython

        if IPython.get_ipython() is not None:
            return True
    return False


def password_grant(http_client, auth_endpoint, provider, username, password):
    form_data = {
        "grant_type": "password",
        "username": username,
        "password": password,
    }
    for attempt in retry_context():
        with attempt:
            token_response = http_client.post(auth_endpoint, data=form_data, auth=None)
            handle_error(token_response)
    return token_response.json()


def device_code_grant(http_client, auth_endpoint):
    for attempt in retry_context():
        with attempt:
            verification_response = http_client.post(auth_endpoint, json={}, auth=None)
            handle_error(verification_response)
    verification = verification_response.json()
    authorization_uri = verification["authorization_uri"]
    print(
        f"""
You have {int(verification['expires_in']) // 60} minutes to visit this URL

{authorization_uri}

and enter the code:

{verification['user_code']}

"""
    )
    import webbrowser

    webbrowser.open(authorization_uri)
    deadline = verification["expires_in"] + time.monotonic()
    print("Waiting...", end="", flush=True)
    while True:
        time.sleep(verification["interval"])
        if time.monotonic() > deadline:
            raise Exception("Deadline expired.")
        for attempt in retry_context():
            with attempt:
                # Intentionally do not wrap this in handle_error(...).
                # Check status codes manually below.
                access_response = http_client.post(
                    verification["verification_uri"],
                    json={
                        "device_code": verification["device_code"],
                        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                    },
                    auth=None,
                )
                if (access_response.status_code == httpx.codes.BAD_REQUEST) and (
                    access_response.json()["detail"]["error"] == "authorization_pending"
                ):
                    print(".", end="", flush=True)
                    continue
                handle_error(access_response)
        print("")
        break
    tokens = access_response.json()
    return tokens



import math
import string
import warnings

import dask.array
import numpy
import pandas.testing
import pytest
import uvicorn
import zarr
from fsspec.implementations.http import HTTPFileSystem
from httpx import ASGITransport, AsyncClient
from starlette.status import HTTP_200_OK, HTTP_401_UNAUTHORIZED, HTTP_404_NOT_FOUND

from ..adapters.array import ArrayAdapter
from ..adapters.dataframe import DataFrameAdapter
from ..adapters.mapping import MapAdapter
from ..adapters.zarr import ZARR_LIB_V2
from ..server.app import build_app
from .utils import Server

url_prefixes = ["/zarr/v2"] if ZARR_LIB_V2 else ["/zarr/v2", "/zarr/v3"]
pytestmark = pytest.mark.parametrize("prefix", url_prefixes)

rng = numpy.random.default_rng(seed=42)
array_cases = {
    "dtype_b": (numpy.arange(1, 10) % 2).astype("b"),
    "dtype_i": rng.integers(-10, 10, size=10, dtype="i"),
    "dtype_uint8": rng.integers(1, 10, size=10, dtype="uint8"),
    "dtype_uint16": rng.integers(1, 10, size=10, dtype="uint16"),
    "dtype_uint64": rng.integers(1, 10, size=10, dtype="uint64"),
    "dtype_f": rng.random(10, dtype="f"),
    "dtype_c": (rng.random(10) + rng.random(10) * 1j).astype("c8"),
    "dtype_S": numpy.array([letter * 3 for letter in string.ascii_letters], dtype="S3"),
    "dtype_U": numpy.array([letter * 3 for letter in string.ascii_letters], dtype="U3"),
    "dtype_m": numpy.array(
        ["2007-07-13", "2006-01-13", "2010-08-13"], dtype="datetime64"
    )
    - numpy.datetime64("2008-01-01"),
    "dtype_M": numpy.array(
        ["2007-07-13", "2006-01-13", "2010-08-13"], dtype="datetime64"
    ),
    "dtype_struct": numpy.array(
        [("Rex", 9, 81.0), ("Fido", 3, 27.0), ("Spot", 5, 45.0)],
        dtype=[("name", "U10"), ("age", "i4"), ("weight", "f4")],
    ),
    "random_2d": rng.random((10, 10)),
}
# TODO bitfield "t", void "v", and object "O" (which is not supported by default)
scalar_cases = {
    k: numpy.array(v[0], dtype=v.dtype)
    for k, v in array_cases.items()
    if k.startswith("dtype_")
}
for v in scalar_cases.values():
    assert v.shape == ()
array_tree = MapAdapter({k: ArrayAdapter.from_array(v) for k, v in array_cases.items()})
scalar_tree = MapAdapter(
    {k: ArrayAdapter.from_array(v) for k, v in scalar_cases.items()}
)

cube_cases = {
    "tiny_cube": rng.random((10, 10, 10)),
    "tiny_hypercube": rng.random((10, 10, 10, 10, 10)),
}
cube_tree = MapAdapter({k: ArrayAdapter.from_array(v) for k, v in cube_cases.items()})
arr_with_inf = numpy.array([0, 1, numpy.nan, -numpy.inf, numpy.inf])
inf_tree = MapAdapter(
    {
        "example": ArrayAdapter.from_array(
            arr_with_inf,
            metadata={"infinity": math.inf, "-infinity": -math.inf, "nan": numpy.nan},
        )
    },
    metadata={"infinity": math.inf, "-infinity": -math.inf, "nan": numpy.nan},
)
arr_with_zero_dim = numpy.array([]).reshape((0, 100, 1, 10))
# Suppress RuntimeWarning: divide by zero encountered in true_divide from dask.array.core.
with warnings.catch_warnings():
    zero_tree = MapAdapter(
        {
            "example": ArrayAdapter.from_array(
                dask.array.from_array(arr_with_zero_dim, chunks=arr_with_zero_dim.shape)
            )
        }
    )
df = pandas.DataFrame(
    {
        "x": rng.random(size=10, dtype="float64"),
        "y": rng.integers(10, size=10, dtype="uint"),
        "z": rng.integers(-10, 10, size=10, dtype="int64"),
    }
)
table_tree = MapAdapter(
    {
        # a dataframe divided into three partitions
        "divided": DataFrameAdapter.from_pandas(df, npartitions=3),
        # a dataframe with just one partition
        "single": DataFrameAdapter.from_pandas(df, npartitions=1),
    }
)

tree = MapAdapter(
    {
        "nested": MapAdapter({"array": array_tree, "cube": cube_tree}),
        "inf": inf_tree,
        "scalar": scalar_tree,
        "zero": zero_tree,
        "table": table_tree,
        "random_2d": array_tree["random_2d"],
    }
)


def traverse_tree(tree, parent="", result=None):
    result = result or {}
    for key, val in tree.items():
        if isinstance(val, ArrayAdapter):
            result.update({f"{parent}/{key}": "array"})
        elif isinstance(val, DataFrameAdapter):
            result.update({f"{parent}/{key}": "group"})
            for col, _ in val.items():
                result.update({f"{parent}/{key}/{col}": "array"})
        else:
            result.update({f"{parent}/{key}": "group"})
            traverse_tree(val, parent=f"{parent}/{key}", result=result)
    return result


@pytest.fixture(scope="module")
def app():
    app = build_app(tree, authentication={"single_user_api_key": "secret"})
    return app


@pytest.fixture(scope="module")
def server_url(app):
    config = uvicorn.Config(app, host="127.0.0.1", port=0, log_level="info")
    server = Server(config)
    with server.run_in_thread() as url:
        yield url


@pytest.fixture(scope="module")
def fs():
    headers = {"Authorization": "Apikey secret", "Content-Type": "application/json"}
    fs = HTTPFileSystem(
        client_kwargs={"headers": headers}, asynchronous=not ZARR_LIB_V2
    )
    return fs


@pytest.mark.parametrize("path", ["", "/", "/nested", "/table/single"])
@pytest.mark.asyncio
async def test_zarr_group_routes(prefix, path, app):
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Apikey secret"},
        follow_redirects=True,
    ) as client:
        response = await client.get(prefix + path)
        assert response.status_code == HTTP_200_OK

        response = await client.get(prefix + path + "/.zarray")
        assert response.status_code == HTTP_404_NOT_FOUND

        response = await client.get(prefix + path + "/.zgroup")
        assert (
            response.status_code == HTTP_200_OK
            if prefix == "/zarr/v2"
            else HTTP_404_NOT_FOUND
        )


@pytest.mark.parametrize("path", ["/nested/cube/tiny_cube", "/table/single/x"])
@pytest.mark.asyncio
async def test_zarr_array_routes(prefix, path, app):
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Apikey secret"},
        follow_redirects=True,
    ) as client:
        response = await client.get(prefix + path)
        assert response.status_code == HTTP_200_OK

        response = await client.get(prefix + path + "/.zgroup")
        assert response.status_code == HTTP_404_NOT_FOUND

        if prefix == "/zarr/v2":
            response = await client.get(prefix + path + "/.zarray")
            assert response.status_code == HTTP_200_OK

            ndim = len(response.json().get("shape"))
            indx = ".".join(["0"] * max(ndim, 0))
            response = await client.get(prefix + path + f"/{indx}")
            assert response.status_code == HTTP_200_OK


@pytest.mark.parametrize(
    "path",
    [
        "",
        "/",
        "/nested",
        "/table/single",
        "/nested/cube/tiny_cube",
        "/table/single/x",
    ],
)
@pytest.mark.asyncio
async def test_authentication(prefix, path, app):
    async with AsyncClient(
        transport=ASGITransport(app=app),
        base_url="http://test",
        headers={"Authorization": "Apikey not-secret"},
        follow_redirects=True,
    ) as client:
        response = await client.get(prefix + path)
        assert response.status_code == HTTP_401_UNAUTHORIZED

        response = await client.get(prefix + path + "/.zarray")
        assert response.status_code == HTTP_401_UNAUTHORIZED

        response = await client.get(prefix + path + "/.zgroup")
        assert response.status_code == HTTP_401_UNAUTHORIZED


def test_zarr_integration(server_url, fs, prefix):
    url = f"{server_url}{prefix}"
    grp = zarr.open(fs.get_mapper(url), mode="r")

    assert grp.store.fs == fs
    assert set(grp.keys()) == set(tree.keys())
    assert len(set(grp.group_keys())) == 5
    assert len(set(grp.array_keys())) == 1


@pytest.mark.parametrize(
    "suffix, path",
    [
        ("", "random_2d"),
        ("", "nested/array/random_2d"),
        ("nested", "array/random_2d"),
        ("nested/array", "random_2d"),
        ("nested/array/random_2d", ""),
    ],
)
@pytest.mark.parametrize("slash", ["", "/"])
def test_zarr_groups(prefix, path, suffix, slash, server_url, fs):
    expected = array_cases["random_2d"]
    url = f"{server_url}{prefix}/{suffix}{slash}"
    arr = zarr.open(fs.get_mapper(url), mode="r")
    if path:
        arr = arr[path]
    assert numpy.array_equal(arr[...], expected)


@pytest.mark.parametrize("kind", list(array_cases.keys()))
def test_array_dtypes(kind, prefix, server_url, fs):
    expected = array_cases[kind]
    url = f"{server_url}{prefix}/nested/array"
    grp = zarr.open(fs.get_mapper(url), mode="r")
    actual = grp[kind][...]
    assert numpy.array_equal(actual, expected)


@pytest.mark.parametrize("kind", list(scalar_cases))
def test_scalar_dtypes(kind, prefix, server_url, fs):
    expected = scalar_cases[kind]
    url = f"{server_url}{prefix}/scalar"
    grp = zarr.open(fs.get_mapper(url), mode="r")
    actual = grp[kind][...]
    assert numpy.array_equal(actual, expected)


@pytest.mark.parametrize("kind", list(cube_cases))
def test_cube_cases(kind, prefix, server_url, fs):
    expected = cube_cases[kind]
    url = f"{server_url}{prefix}/nested/cube"
    grp = zarr.open(fs.get_mapper(url), mode="r")
    actual = grp[kind][...]
    assert numpy.array_equal(actual, expected)


def test_infinity(prefix, server_url, fs):
    url = f"{server_url}{prefix}/inf/example"
    actual = zarr.open(fs.get_mapper(url), mode="r")[...]
    mask = numpy.isnan(arr_with_inf)
    assert numpy.array_equal(actual[~mask], arr_with_inf[~mask])
    assert numpy.isnan(actual[mask]).all()


def test_shape_with_zero(prefix, server_url, fs):
    url = f"{server_url}{prefix}/zero/example"
    actual = zarr.open(fs.get_mapper(url), mode="r")[...]
    assert numpy.array_equal(actual, arr_with_zero_dim)


def test_dataframe_group(prefix, server_url, fs):
    url = f"{server_url}{prefix}/table"
    grp = zarr.open(fs.get_mapper(url), mode="r")
    assert set(grp.keys()) == set(table_tree.keys())

    for key in grp.keys():
        for col in grp[key].keys():
            actual = grp[key][col][...]
            expected = df[col]
            assert numpy.array_equal(actual, expected)


@pytest.mark.parametrize("key", list(table_tree.keys()))
def test_dataframe_single(key, prefix, server_url, fs):
    url = f"{server_url}{prefix}/table/{key}"
    grp = zarr.open(fs.get_mapper(url), mode="r")

    for col in df.columns:
        actual = grp[col][...]
        expected = df[col]
        assert numpy.array_equal(actual, expected)


@pytest.mark.parametrize("key", list(table_tree.keys()))
def test_dataframe_column(key, prefix, server_url, fs):
    for col in df.columns:
        url = f"{server_url}{prefix}/table/{key}/{col}"
        arr = zarr.open(fs.get_mapper(url), mode="r")
        actual = arr[...]
        expected = df[col]
        assert numpy.array_equal(actual, expected)


def test_writing_not_implemented(prefix, server_url, fs):
    url = f"{server_url}{prefix}/nested/array"

    # with pytest.raises(NotImplementedError):
    #     zarr.open(fs.get_mapper(url), mode="w")

    with pytest.raises(zarr.errors.ReadOnlyError if ZARR_LIB_V2 else ValueError):
        grp = zarr.open(fs.get_mapper(url), mode="r")
        grp["random_2d"][0, 0] = 0.0



from pathlib import Path

from tiled.access_control.access_tags import AccessTagsCompiler
from tiled.access_control.scopes import ALL_SCOPES


def group_parser(groupname):
    return {
        "group_A": ["alice", "bob"],
        "admins": ["cara"],
    }[groupname]


def main():
    file_directory = Path(__file__).resolve().parent

    access_tags_compiler = AccessTagsCompiler(
        ALL_SCOPES,
        Path(file_directory, "tag_definitions.yml"),
        {"uri": f"file:{file_directory}/compiled_tags.sqlite"},
        group_parser,
    )

    access_tags_compiler.load_tag_config()
    access_tags_compiler.compile()
    access_tags_compiler.connection.close()


if __name__ == "__main__":
    main()



import os
from typing import Any, Callable, Optional

import cachetools  # type: ignore

# When items are evicted from the cache, a hard reference is dropped, freeing
# the resource to be closed by the garbage collector if there are no other
# extant hard references. Items are evicted if:
#
# - They have been in the cache for a _total_ of more than a given time.
#   (Accessing an item does not reset this time.)
# - The cache is at capacity and this item is the least recently used item.
#
# The "size" is measured in cached items; that is, each item in the cache has
# size 1.
DEFAULT_MAX_SIZE = int(os.getenv("TILED_RESOURCE_CACHE_MAX_SIZE", "1024"))
DEFAULT_TIME_TO_USE_SECONDS = float(os.getenv("TILED_RESOURCE_CACHE_TTU", "60."))


def get_resource_cache() -> cachetools.Cache:
    "Return resource cache, a process-global Cache."
    return _cache


def set_resource_cache(cache: cachetools.Cache) -> None:
    """
    Set the resource cache, a process-global Cache.
    Parameters
    ----------
    cache :

    Returns
    -------

    """
    global _cache
    _cache = cache


def default_ttu(_key: str, value: Any, now: float) -> float:
    """
    Retain cached items for at most `DEFAULT_TIME_TO_USE_SECONDS` seconds (60s, by default).

    Parameters
    ----------
    _key :
    value :
    now :

    Returns
    -------

    """
    return DEFAULT_TIME_TO_USE_SECONDS + now


def default_resource_cache() -> cachetools.TLRUCache:
    "Create a new instance of the default resource cache."
    return cachetools.TLRUCache(DEFAULT_MAX_SIZE, default_ttu)


def with_resource_cache(
    cache_key: Any,
    factory: Callable[..., Any],
    *args: Any,
    _resource_cache: Optional[cachetools.Cache] = None,
    **kwargs: Any,
) -> Any:
    """
    Use value from cache or, if not present, call `factory(*args, **kwargs)` and cache result.

    This uses a globally configured resource cache by default. For testing and
    debugging, a cache may be passed to the parameter _resource_cache.

    Parameters
    ----------
    cache_key :
    factory :
    args :
    _resource_cache :
    kwargs :

    Returns
    -------

    """
    if _resource_cache is None:
        cache = get_resource_cache()
    else:
        cache = _resource_cache
    # Return cached value if found.
    value = cache.get(cache_key)
    if value is not None:
        return value
    # Generate value and offer it to the cache.
    value = factory(*args, **kwargs)
    if cache.maxsize:  # handle size 0 cache
        cache[cache_key] = value
    return value


_cache: cachetools.Cache = default_resource_cache()



from typing import Callable, cast

import numpy as np
import pyarrow as pa
import pytest

from tiled._tests.adapters.test_sql import adapter_duckdb_many_partitions  # noqa: F401
from tiled._tests.adapters.test_sql import adapter_duckdb_one_partition  # noqa: F401
from tiled._tests.adapters.test_sql import adapter_psql_many_partitions  # noqa: F401
from tiled._tests.adapters.test_sql import adapter_psql_one_partition  # noqa: F401
from tiled._tests.adapters.test_sql import assert_same_rows
from tiled.adapters.sql import SQLAdapter
from tiled.storage import SQLStorage, parse_storage, register_storage
from tiled.structures.core import StructureFamily
from tiled.structures.data_source import DataSource, Management
from tiled.structures.table import TableStructure

rng = np.random.default_rng(42)

names = ["i0", "i1", "i2", "i3", "f4", "f5"]
batch_size = 5
data0 = [
    pa.array(
        [rng.integers(-100, 100, size=10, dtype=np.int8) for _ in range(batch_size)]
    ),
    pa.array(
        [rng.integers(-100, 100, size=11, dtype=np.int16) for _ in range(batch_size)]
    ),
    pa.array(
        [rng.integers(-100, 100, size=12, dtype=np.int32) for _ in range(batch_size)]
    ),
    pa.array(
        [rng.integers(-100, 100, size=13, dtype=np.int64) for _ in range(batch_size)]
    ),
    pa.array([rng.random(size=14, dtype=np.float32) for _ in range(batch_size)]),
    pa.array([rng.random(size=15, dtype=np.float64) for _ in range(batch_size)]),
]
batch_size = 8
data1 = [
    pa.array(
        [rng.integers(-100, 100, size=10, dtype=np.int8) for _ in range(batch_size)]
    ),
    pa.array(
        [rng.integers(-100, 100, size=11, dtype=np.int16) for _ in range(batch_size)]
    ),
    pa.array(
        [rng.integers(-100, 100, size=12, dtype=np.int32) for _ in range(batch_size)]
    ),
    pa.array(
        [rng.integers(-100, 100, size=13, dtype=np.int64) for _ in range(batch_size)]
    ),
    pa.array([rng.random(size=14, dtype=np.float32) for _ in range(batch_size)]),
    pa.array([rng.random(size=15, dtype=np.float64) for _ in range(batch_size)]),
]
batch_size = 3
data2 = [
    pa.array(
        [rng.integers(-100, 100, size=10, dtype=np.int8) for _ in range(batch_size)]
    ),
    pa.array(
        [rng.integers(-100, 100, size=11, dtype=np.int16) for _ in range(batch_size)]
    ),
    pa.array(
        [rng.integers(-100, 100, size=12, dtype=np.int32) for _ in range(batch_size)]
    ),
    pa.array(
        [rng.integers(-100, 100, size=13, dtype=np.int64) for _ in range(batch_size)]
    ),
    pa.array([rng.random(size=14, dtype=np.float32) for _ in range(batch_size)]),
    pa.array([rng.random(size=15, dtype=np.float64) for _ in range(batch_size)]),
]

batch0 = pa.record_batch(data0, names=names)
batch1 = pa.record_batch(data1, names=names)
batch2 = pa.record_batch(data2, names=names)


@pytest.fixture
def data_source_from_init_storage() -> Callable[[str, int], DataSource[TableStructure]]:
    def _data_source_from_init_storage(
        data_uri: str, num_partitions: int
    ) -> DataSource[TableStructure]:
        table = pa.Table.from_arrays(data0, names)
        structure = TableStructure.from_arrow_table(table, npartitions=num_partitions)
        data_source = DataSource(
            management=Management.writable,
            mimetype="application/x-tiled-sql-table",
            structure_family=StructureFamily.table,
            structure=structure,
            assets=[],
        )

        storage = cast(SQLStorage, parse_storage(data_uri))
        register_storage(storage)
        return SQLAdapter.init_storage(data_source=data_source, storage=storage)

    return _data_source_from_init_storage


@pytest.mark.parametrize(
    "adapter_name", [("adapter_duckdb_one_partition"), ("adapter_psql_one_partition")]
)
def test_write_read_one_batch_one_part(
    adapter_name: str, request: pytest.FixtureRequest
) -> None:
    # get adapter from fixture
    adapter: SQLAdapter = request.getfixturevalue(adapter_name)

    # test appending and reading a table as a whole
    test_table = pa.Table.from_arrays(data0, names)

    adapter.append_partition(batch0, 0)
    result_read = adapter.read()
    assert test_table == pa.Table.from_pandas(result_read)

    # test appending and reading a partition in a table
    result_read_partition = adapter.read_partition(0)
    assert test_table == pa.Table.from_pandas(result_read_partition)


@pytest.mark.parametrize(
    "adapter_name", [("adapter_duckdb_one_partition"), ("adapter_psql_one_partition")]
)
def test_write_read_list_batch_one_part(
    adapter_name: str, request: pytest.FixtureRequest
) -> None:
    # get adapter from fixture
    adapter: SQLAdapter = request.getfixturevalue(adapter_name)

    test_table = pa.Table.from_batches([batch0, batch1, batch2])
    # test appending a list of batches to a table and read as a whole
    adapter.append_partition([batch0, batch1, batch2], 0)
    result_read = adapter.read()

    assert test_table == pa.Table.from_pandas(result_read)

    # test appending and reading a partition in a table
    result_read_partition = adapter.read_partition(0)

    assert test_table == pa.Table.from_pandas(result_read_partition)

    # test appending few more times done correctly
    test_table = pa.Table.from_batches(
        [batch0, batch1, batch2, batch2, batch0, batch1, batch1, batch2, batch0]
    )
    adapter.append_partition([batch2, batch0, batch1], 0)
    adapter.append_partition([batch1, batch2, batch0], 0)
    result_read = adapter.read()

    assert test_table == pa.Table.from_pandas(result_read)

    # test appending a few times and reading done correctly
    result_read_partition = adapter.read_partition(0)

    assert test_table == pa.Table.from_pandas(result_read_partition)


@pytest.mark.parametrize(
    "adapter_name",
    [("adapter_duckdb_many_partitions"), ("adapter_psql_many_partitions")],
)
def test_append_single_partition(
    adapter_name: str, request: pytest.FixtureRequest
) -> None:
    # get adapter from fixture
    adapter: SQLAdapter = request.getfixturevalue(adapter_name)

    # test writing an entire pyarrow table to a single partition
    table = pa.Table.from_batches([batch0, batch1, batch2])
    adapter.append_partition(table, 0)

    result_read = adapter.read()
    assert table == pa.Table.from_pandas(result_read)

    # test reading a specific partition
    result_read_partition = adapter.read_partition(0)
    assert table == pa.Table.from_pandas(result_read_partition)


@pytest.mark.parametrize("adapter_name", [("adapter_psql_many_partitions")])
@pytest.mark.parametrize("field", names)
def test_write_read_one_batch_many_part(
    adapter_name: str, request: pytest.FixtureRequest, field: str
) -> None:
    # get adapter from fixture
    adapter: SQLAdapter = request.getfixturevalue(adapter_name)

    # test writing to many partitions and reading it whole
    adapter.append_partition(batch0, 0)
    adapter.append_partition(batch1, 1)
    adapter.append_partition(batch2, 2)

    result_read = adapter.read()

    assert pa.Table.from_batches([batch0, batch1, batch2]) == pa.Table.from_pandas(
        result_read
    )

    # test reading a specific partition
    result_read_partition = adapter.read_partition(0)
    assert pa.Table.from_arrays(data0, names) == pa.Table.from_pandas(
        result_read_partition
    )

    result_read_partition = adapter.read_partition(1)
    assert pa.Table.from_arrays(data1, names) == pa.Table.from_pandas(
        result_read_partition
    )

    result_read_partition = adapter.read_partition(2)
    assert pa.Table.from_arrays(data2, names) == pa.Table.from_pandas(
        result_read_partition
    )

    # test appending a few times and reading done correctly
    adapter.append_partition(batch0, 1)
    adapter.append_partition(batch1, 2)
    adapter.append_partition(batch2, 0)

    result_read = adapter.read()

    # Check that each partition matches
    assert_same_rows(
        pa.Table.from_batches([batch0, batch2]),
        pa.Table.from_pandas(adapter.read_partition(0)),
    )
    assert_same_rows(
        pa.Table.from_batches([batch1, batch0]),
        pa.Table.from_pandas(adapter.read_partition(1)),
    )
    assert_same_rows(
        pa.Table.from_batches([batch2, batch1]),
        pa.Table.from_pandas(adapter.read_partition(2)),
    )
    assert_same_rows(
        pa.Table.from_batches([batch0, batch2, batch1, batch0, batch2, batch1]),
        pa.Table.from_pandas(result_read),
    )

    # read a specific field
    result_read = adapter.read_partition(0, fields=[field])
    field_index = names.index(field)
    assert np.array_equal(
        [*data0[field_index].tolist(), *data2[field_index].tolist()],
        result_read[field].tolist(),
    )
    result_read = adapter.read_partition(1, fields=[field])
    assert np.array_equal(
        [*data1[field_index].tolist(), *data0[field_index].tolist()],
        result_read[field].tolist(),
    )
    result_read = adapter.read_partition(2, fields=[field])
    assert np.array_equal(
        [*data2[field_index].tolist(), *data1[field_index].tolist()],
        result_read[field].tolist(),
    )



import dataclasses
import platform
import random
from pathlib import Path

import h5py
import numpy
import pytest
import tifffile
import yaml
from starlette.status import HTTP_415_UNSUPPORTED_MEDIA_TYPE

from ..adapters.hdf5 import HDF5Adapter
from ..adapters.tiff import TiffAdapter
from ..adapters.utils import init_adapter_from_catalog
from ..catalog import in_memory
from ..client import Context, from_context
from ..client.register import (
    Settings,
    group_image_sequences,
    identity,
    register,
    register_image_sequence,
    skip_all,
    strip_suffixes,
)
from ..examples.generate_files import data, df1, generate_files
from ..server.app import build_app
from ..structures.array import ArrayStructure
from ..structures.data_source import Asset, DataSource, Management
from ..utils import ensure_uri, path_from_uri
from .utils import fail_with_status_code


@pytest.fixture
def example_data_dir(tmpdir_factory):
    """
    Generate a temporary directory with example files.

    The tmpdir_factory fixture ensures that this directory is cleaned up at test exit.
    """
    tmpdir = tmpdir_factory.mktemp("example_files")
    generate_files(tmpdir)
    return tmpdir


@pytest.mark.xfail(
    platform.system() == "Windows",
    reason="file cannot be removed while being used",
    raises=PermissionError,
)
@pytest.mark.asyncio
async def test_collision(example_data_dir, tmpdir):
    """Test that files which produce key collisions are ignored until the collision is resolved."""
    # Add a.tiff which will collide with a.tif.
    p = Path(example_data_dir, "a.tiff")
    tifffile.imwrite(str(p), data)

    catalog = in_memory(writable_storage=str(tmpdir))
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        await register(client, example_data_dir)

        # And omits the colliding entries.
        assert "a" not in client

        # Resolve the collision.
        p.unlink()

        # Re-run registration; entry should be there now.
        await register(client, example_data_dir)
        assert "a" in client


@pytest.mark.parametrize(
    ("filename", "expected"), [("a.txt", "a"), ("a.tar.gz", "a"), ("a", "a")]
)
def test_strip_suffixes(filename, expected):
    actual = strip_suffixes(filename)
    assert actual == expected


@pytest.mark.asyncio
async def test_same_filename_separate_directory(tmpdir):
    "Two files with the same name in separate directories should not collide."
    Path(tmpdir, "one").mkdir()
    Path(tmpdir, "two").mkdir()
    df1.to_csv(Path(tmpdir, "one", "a.csv"))
    df1.to_csv(Path(tmpdir, "two", "a.csv"))
    catalog = in_memory(writable_storage=str(tmpdir))
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        await register(client, tmpdir)
        assert "a" in client["one"]
        assert "a" in client["two"]


@pytest.mark.asyncio
async def test_mimetype_detection_hook(tmpdir):
    content = "a, b, c\n1, 2 ,3\n4, 5, 6\n"
    with open(Path(tmpdir / "a0"), "w") as file:
        file.write(content)
    with open(Path(tmpdir / "b0"), "w") as file:
        file.write(content)
    with open(Path(tmpdir / "c.csv"), "w") as file:
        file.write(content)
    with open(Path(tmpdir / "a.0.asfwoeijviojefeiofw"), "w") as file:
        file.write(content)
    with open(Path(tmpdir / "b.0.asfwoeijviojefeiofw"), "w") as file:
        file.write(content)

    def detect_mimetype(path, mimetype):
        filename = Path(path).name
        # If detection based on file extension worked,
        # we should get that in the mimetype. Otherwise,
        # mimetype should be None.
        if filename.endswith(".csv"):
            assert mimetype == "text/csv"
        else:
            assert mimetype is None
        if filename.startswith("a"):
            return "text/csv"
        return mimetype

    catalog = in_memory(writable_storage=str(tmpdir))
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        await register(
            client,
            tmpdir,
            mimetype_detection_hook=detect_mimetype,
            key_from_filename=identity,
        )
        assert set(client) == {"a0", "a.0.asfwoeijviojefeiofw", "c.csv"}


@pytest.mark.asyncio
async def test_skip_all_in_combination(tmpdir):
    "Using skip_all should override defaults, but not hinder other walkers"
    df1.to_csv(Path(tmpdir, "a.csv"))
    Path(tmpdir, "one").mkdir()
    df1.to_csv(Path(tmpdir, "one", "a.csv"))

    for i in range(2):
        tifffile.imwrite(Path(tmpdir, "one", f"image{i:05}.tif"), data)

    catalog = in_memory(writable_storage=str(tmpdir))
    # By default, both file and tiff sequence are registered.
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        await register(client, tmpdir)
        assert "a" in client
        assert "a" in client["one"]
        assert "image" in client["one"]

    # With skip_all, directories and tiff sequence are registered, but individual files are not
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        await register(client, tmpdir, walkers=[group_image_sequences, skip_all])
        assert list(client) == ["one"]
        assert "image" in client["one"]


@pytest.mark.asyncio
async def test_tiff_seq_custom_sorting(tmpdir):
    "Register images that are not in alphanumeric order."
    N = 10
    ordering = list(range(N))
    random.Random(0).shuffle(ordering)
    files = []
    for i in ordering:
        file = Path(tmpdir, f"image{i:05}.tif")
        files.append(file)
        # data is a block of ones
        tifffile.imwrite(file, i * data)

    settings = Settings.init()
    catalog = in_memory(writable_storage=str(tmpdir))
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        await register_image_sequence(
            client,
            "image",
            files,
            settings,
        )
        # We are being a bit clever here.
        # Each image in this image series has pixels with a constant value, and
        # that value matches the image's position in the sequence enumerated by
        # `ordering`. We pick out one pixel and check that its value matches
        # the corresponding value in `ordering`.
        actual = list(client["image"][:, 0, 0])
        assert actual == ordering


@pytest.mark.asyncio
async def test_image_file_with_sidecar_metadata_file(tmpdir):
    "Create one Node from two different types of files."
    MIMETYPE = "multipart/related;type=application/x-tiff-with-yaml"
    image_filepath = Path(tmpdir, "image.tif")
    tifffile.imwrite(image_filepath, data)
    metadata_filepath = Path(tmpdir, "metadata.yml")
    metadata = {"test_key": 3.0}
    with open(metadata_filepath, "w") as file:
        yaml.dump(metadata, file)

    class TiffAdapterWithSidecar(TiffAdapter):
        def __init__(self, image_uri, metadata_uri, metadata=None, **kwargs):
            with open(path_from_uri(metadata_uri)) as file:
                metadata = yaml.safe_load(file)

            super().__init__(image_uri, metadata=metadata, **kwargs)

        @classmethod
        def from_catalog(
            cls,
            data_source,
            node,
            /,
            **kwargs,
        ):
            return init_adapter_from_catalog(cls, data_source, node, **kwargs)

    catalog = in_memory(
        writable_storage=str(tmpdir),
        adapters_by_mimetype={MIMETYPE: TiffAdapterWithSidecar},
    )
    with Context.from_app(build_app(catalog)) as context:
        adapter = TiffAdapterWithSidecar(
            ensure_uri(image_filepath), ensure_uri(metadata_filepath)
        )
        client = from_context(context)
        client.new(
            key="image",
            structure_family=adapter.structure_family,
            metadata=dict(adapter.metadata()),
            specs=adapter.specs,
            data_sources=[
                DataSource(
                    structure_family=adapter.structure_family,
                    mimetype=MIMETYPE,
                    structure=dataclasses.asdict(adapter.structure()),
                    parameters={},
                    management=Management.external,
                    assets=[
                        Asset(
                            data_uri=ensure_uri(metadata_filepath),
                            is_directory=False,
                            parameter="metadata_uri",
                        ),
                        Asset(
                            data_uri=ensure_uri(image_filepath),
                            is_directory=False,
                            parameter="image_uri",
                        ),
                    ],
                )
            ],
        )
        assert numpy.array_equal(data, client["image"][:])
        assert client["image"].metadata["test_key"] == 3.0


@pytest.mark.asyncio
async def test_hdf5_virtual_datasets(tmpdir):
    # A virtual database comprises one master file and N data files. The master
    # file must be handed to the Adapter for opening. The data files are not
    # handled directly by the Adapter but they still ought to be tracked as
    # Assets for purposes of data movement, accounting for data size, etc.
    # This is why they are Assets with parameter=NULL/None, Assets not used
    # directly by the Adapter.

    # One could do one-dataset-per-directory. But like TIFF series in practice
    # they are often mixed, so we address that general case and track them at
    # the per-file level.

    # Contrast this to Zarr, where the files involves are always bundled by
    # directory. We track Zarr at the directory level.

    layout = h5py.VirtualLayout(shape=(4, 100), dtype="i4")

    data_filepaths = []
    for n in range(1, 5):
        filepath = Path(tmpdir, f"{n}.h5")
        data_filepaths.append(filepath)
        vsource = h5py.VirtualSource(filepath, "data", shape=(100,))
        layout[n - 1] = vsource

    # Add virtual dataset to output file
    filepath = Path(tmpdir, "VDS.h5")
    with h5py.File(filepath, "w", libver="latest") as file:
        file.create_virtual_dataset("data", layout, fillvalue=-5)

    assets = [
        Asset(
            data_uri=str(ensure_uri(str(fp))),
            is_directory=False,
            parameter=None,  # an indirect dependency
        )
        for fp in data_filepaths
    ]
    assets.append(
        Asset(
            data_uri=ensure_uri(filepath),
            is_directory=False,
            parameter="data_uris",
        )
    )
    catalog = in_memory(writable_storage=str(tmpdir))
    with Context.from_app(build_app(catalog)) as context:
        adapter = HDF5Adapter.from_uris(ensure_uri(filepath))
        client = from_context(context)
        client.new(
            key="VDS",
            structure_family=adapter.structure_family,
            metadata=dict(adapter.metadata()),
            specs=adapter.specs,
            data_sources=[
                DataSource(
                    structure_family=adapter.structure_family,
                    mimetype="application/x-hdf5",
                    structure=None,
                    parameters={},
                    management=Management.external,
                    assets=assets,
                )
            ],
        )
        client["VDS"]["data"][:]


def test_unknown_mimetype(tmpdir):
    catalog = in_memory(writable_storage=str(tmpdir))
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        asset = Asset(
            data_uri=ensure_uri(tmpdir / "test.does_not_exist"),
            is_directory=False,
            parameter="test",
        )
        with fail_with_status_code(HTTP_415_UNSUPPORTED_MEDIA_TYPE):
            client.new(
                key="x",
                structure_family="array",
                metadata={},
                specs=[],
                data_sources=[
                    DataSource(
                        structure_family="array",
                        mimetype="application/x-does-not-exist",
                        structure=None,
                        parameters={},
                        management=Management.external,
                        assets=[asset],
                    )
                ],
            )


def test_one_asset_two_data_sources(tmpdir):
    catalog = in_memory(writable_storage=str(tmpdir))
    with Context.from_app(build_app(catalog)) as context:
        client = from_context(context)
        asset = Asset(
            data_uri=ensure_uri(tmpdir / "test.tiff"),
            is_directory=False,
            parameter="data_uri",
            num=None,
        )
        for key in ["x", "y"]:
            client.new(
                key=key,
                structure_family="array",
                metadata={},
                specs=[],
                data_sources=[
                    DataSource(
                        structure_family="array",
                        mimetype="image/tiff",
                        structure=ArrayStructure.from_array(numpy.empty((5, 7))),
                        parameters={},
                        management=Management.external,
                        assets=[asset],
                    )
                ],
            )



import enum
from dataclasses import dataclass, field
from typing import Optional, Tuple, Union

from .array import BuiltinDtype, Endianness, Kind, StructDtype


class SparseLayout(str, enum.Enum):
    # Only COO is currently supported, but this lays a path
    # for adding other layouts like CSC, CSR, etc. in the future.
    COO = "COO"


@dataclass
class COOStructure:
    chunks: Tuple[Tuple[int, ...], ...]  # tuple-of-tuples-of-ints like ((3,), (3,))
    shape: Tuple[int, ...]  # tuple of ints like (3, 3)
    data_type: Optional[Union[BuiltinDtype, StructDtype]] = None
    coord_data_type: Optional[BuiltinDtype] = field(
        default_factory=lambda: BuiltinDtype(
            Endianness("little"), Kind("u"), 8
        )  # numpy 'uint' dtype
    )
    dims: Optional[Tuple[str, ...]] = None  # None or tuple of names like ("x", "y")
    resizable: Union[bool, Tuple[bool, ...]] = False
    layout: SparseLayout = SparseLayout.COO
    # TODO Include fill_value?

    @classmethod
    def from_json(cls, structure):
        data_type = structure.get("data_type", None)
        if data_type is not None and "fields" in data_type:
            data_type = StructDtype.from_json(data_type)
        else:
            data_type = BuiltinDtype.from_json(data_type)
        coord_data_type = structure.get(
            "coord_data_type", {"endianness": "little", "kind": "u", "itemsize": 8}
        )
        return cls(
            data_type=data_type,
            coord_data_type=BuiltinDtype.from_json(coord_data_type),
            chunks=tuple(map(tuple, structure["chunks"])),
            shape=tuple(structure["shape"]),
            dims=structure["dims"],
            resizable=structure.get("resizable", False),
        )


# This may be extended to a Union of structures if more are added.
SparseStructure = COOStructure
